from __future__ import annotations

import dataclasses
import sys
from dataclasses import field
from typing import Any

import jax
import jax.numpy as jnp
import jax.tree_util as jtu
import numpy as np

from pytreeclass.src.decorator_util import dispatch
from pytreeclass.src.tree_viz_util import _format_node_repr, _format_node_str


class static_value:
    def __init__(self, value):
        self.value = value

    def __repr__(self):
        return f"*{_format_node_repr(self.value,0)}"

    def __str__(self):
        return f"*{_format_node_str(self.value,0)}"


def static_field(**kwargs):
    """ignore from pytree computations"""
    return field(**{**kwargs, **{"metadata": {"static": True}}})


def is_treeclass(tree):
    """check if a class is treeclass"""
    return hasattr(tree, "__tree_fields__")


def is_treeclass_leaf_bool(node):
    """assert if treeclass leaf is boolean (for boolen indexing)"""
    if isinstance(node, jnp.ndarray):
        return node.dtype == "bool"
    else:
        return isinstance(node, bool)


def is_treeclass_leaf(tree):
    """assert if a node is treeclass leaf"""
    if is_treeclass(tree):
        fields = tree.__dataclass_fields__.values()

        return is_treeclass(tree) and not any(
            [is_treeclass(tree.__dict__[fi.name]) for fi in fields]
        )
    else:
        return False


def is_treeclass_equal(lhs, rhs):
    """Assert if two treeclasses are equal"""
    lhs_leaves, lhs_treedef = jtu.tree_flatten(lhs)
    rhs_leaves, rhs_treedef = jtu.tree_flatten(rhs)

    def is_node_equal(lhs_node, rhs_node):
        if isinstance(lhs_node, jnp.ndarray) and isinstance(rhs_node, jnp.ndarray):
            return jnp.array_equal(lhs_node, rhs_node)
        else:
            return lhs_node == rhs_node

    return (lhs_treedef == rhs_treedef) and all(
        [is_node_equal(lhs_leaves[i], rhs_leaves[i]) for i in range(len(lhs_leaves))]
    )


def is_excluded(field_item: dataclasses.field, node_item: Any) -> bool:
    """Check if a field is excluded

    Returns:
        bool: boolean if the field should be excluded or not.
    """
    excluded_by_meta = field_item.metadata.get("static", False)
    excluded_by_type = isinstance(node_item, static_value)
    return excluded_by_meta or excluded_by_type


def sequential_tree_shape_eval(tree, array):
    """Evaluate shape propagation of assumed sequential modules"""

    # all dynamic/static leaves
    all_leaves = (
        *tree.__tree_fields__[0].values(),
        *tree.__tree_fields__[1].values(),
    )
    leaves = [leaf for leaf in all_leaves if is_treeclass(leaf)]

    shape = [jax.eval_shape(lambda x: x, array)]
    for leave in leaves:
        shape += [jax.eval_shape(leave, shape[-1])]
    return shape


def _node_count_and_size(node: Any) -> tuple[complex, complex]:
    """Calculate number and size of `trainable` and `non-trainable` parameters

    Args:
        node (Any): treeclass node

    Returns:
        complex: Complex number of (inexact, non-exact) parameters for count/size
    """

    node = node.value if isinstance(node, static_value) else node

    if isinstance(node, (jnp.ndarray, np.ndarray)):
        # inexact(trainable) array
        if jnp.issubdtype(node, jnp.inexact):
            count = complex(int(jnp.array(node.shape).prod()), 0)
            size = complex(int(node.nbytes), 0)

        # exact paramter
        else:
            count = complex(0, int(jnp.array(node.shape).prod()))
            size = complex(0, int(node.nbytes))

    # inexact non-array (array_like)
    elif isinstance(node, (float, complex)):
        count = complex(1, 0)
        size = complex(sys.getsizeof(node), 0)

    # exact non-array
    elif isinstance(node, int):
        count = complex(0, 1)
        size = complex(0, sys.getsizeof(node))

    # exclude others
    else:
        count = complex(0, 0)
        size = complex(0, 0)

    return (count, size)


def _dispatched_tree_map(func, lhs, rhs=None):
    """Slightly different implementation to jtu.tree_map for unary/binary operators broadcasting"""

    @dispatch(argnum=1)
    def _tree_map(lhs, rhs):
        raise NotImplementedError(f"rhs of type {type(rhs)} is not implemented.")

    @_tree_map.register(type(lhs))
    def _(lhs, rhs):
        lhs_leaves, lhs_treedef = jtu.tree_flatten(lhs)
        rhs_leaves, rhs_treedef = jtu.tree_flatten(rhs)

        lhs_leaves = [
            func(lhs_leaf, rhs_leaf) if rhs_leaf is not None else lhs_leaf
            for (lhs_leaf, rhs_leaf) in zip(lhs_leaves, rhs_leaves)
        ]

        return jtu.tree_unflatten(lhs_treedef, lhs_leaves)

    @_tree_map.register(jax.interpreters.partial_eval.DynamicJaxprTracer)
    @_tree_map.register(int)
    @_tree_map.register(float)
    @_tree_map.register(complex)
    @_tree_map.register(bool)
    @_tree_map.register(str)
    def _(lhs, rhs):
        lhs_leaves, lhs_treedef = jtu.tree_flatten(lhs)

        lhs_leaves = [func(leaf, rhs) for leaf in lhs_leaves]

        return jtu.tree_unflatten(lhs_treedef, lhs_leaves)

    @_tree_map.register(type(None))
    def _(lhs, rhs=None):
        lhs_leaves, lhs_treedef = jtu.tree_flatten(lhs)
        lhs_leaves = [func(lhs_node) for lhs_node in lhs_leaves]
        return jtu.tree_unflatten(lhs_treedef, lhs_leaves)

    return _tree_map(lhs, rhs)


def _reduce_count_and_size(leaf):
    """reduce params count and params size of a tree of leaves"""

    def reduce_func(acc, node):
        lhs_count, lhs_size = acc
        rhs_count, rhs_size = _node_count_and_size(node)
        return (lhs_count + rhs_count, lhs_size + rhs_size)

    return jtu.tree_reduce(reduce_func, leaf, (complex(0, 0), complex(0, 0)))


def _freeze_nodes(tree):
    """inplace freezing"""
    if is_treeclass(tree):
        object.__setattr__(tree, "__frozen_fields__", None)
        for kw, leaf in tree.__dataclass_fields__.items():
            _freeze_nodes(tree.__dict__[kw])
    return tree


def _unfreeze_nodes(tree):
    """inplace unfreezing"""
    if is_treeclass(tree):
        if hasattr(tree, "__frozen_fields__"):
            object.__delattr__(tree, "__frozen_fields__")
        for kw, leaf in tree.__dataclass_fields__.items():
            _unfreeze_nodes(tree.__dict__[kw])
    return tree



import jax.numpy as jnp
import pytest

from pytreeclass import treeclass
from pytreeclass.src.tree_util import is_treeclass_equal, static_value


@treeclass
class Test:
    a: float
    b: float
    c: float
    name: str


def test_ops():
    @treeclass
    class Test:
        a: float
        b: float
        c: float
        name: str

    A = Test(10, 20, 30, static_value("A"))
    # binary operations

    assert (A + A) == Test(20, 40, 60, static_value("A"))
    assert (A - A) == Test(0, 0, 0, static_value("A"))
    # assert ((A["a"] + A) | A) == Test(20, 20, 30 ,static_value("A"))
    # assert A.reduce_mean() == jnp.array(60)
    assert abs(A) == A

    @treeclass
    class Test:
        a: int
        b: int
        name: str

    A = Test(-10, 20, static_value("A"))

    # magic ops
    assert abs(A) == Test(10, 20, static_value("A"))
    assert A + A == Test(-20, 40, static_value("A"))
    assert A == A
    assert A // 2 == Test(-5, 10, static_value("A"))
    assert A / 2 == Test(-5.0, 10.0, static_value("A"))
    assert (A > A) == Test(False, False, static_value("A"))
    assert (A >= A) == Test(True, True, static_value("A"))
    assert (A <= A) == Test(True, True, static_value("A"))
    assert -A == Test(10, -20, static_value("A"))
    assert A * A == Test(100, 400, static_value("A"))
    assert A**A == Test((-10) ** (-10), 20**20, static_value("A"))
    assert A - A == Test(0, 0, static_value("A"))

    with pytest.raises(TypeError):
        A + "s"

    with pytest.raises(NotImplementedError):
        A == (1,)

    assert abs(A) == Test(10, 20, static_value("A"))

    # numpy ops
    A = Test(a=jnp.array([-10, -10]), b=1, name="A")


def test_asdict():
    A = Test(10, 20, 30, static_value("A"))
    assert A.asdict() == {"a": 10, "b": 20, "c": 30, "name": "A"}


def test_or():
    @treeclass
    class test:
        a: jnp.ndarray
        b: jnp.ndarray

    x = test(jnp.array([]), jnp.array([4, 5, 6]))
    y = test(jnp.array([1, 2, 3]), jnp.array([]))

    assert is_treeclass_equal((x | y), test(jnp.array([1, 2, 3]), jnp.array([4, 5, 6])))

    # with pytest.raises(ValueError):
    x = test(jnp.array([]), jnp.array([]))
    y = test(jnp.array([1, 2, 3]), None)
    assert is_treeclass_equal(x | y, test(jnp.array([1, 2, 3]), jnp.array([])))



# import jax
import jax.numpy as jnp
import jax.tree_util as jtu

# import numpy as np
import numpy.testing as npt

import pytreeclass as pytc
from pytreeclass import static_field

# import pytest


def test_static():
    @pytc.treeclass
    class Test:
        a: int = 1
        b: int = 2
        c: int = 3

    test = Test()

    assert jtu.tree_leaves(test.at[...].static()) == []
    assert jtu.tree_leaves(test.at[test > 1].static()) == [1]
    assert jtu.tree_leaves(test.at[test < 2].static()) == [2, 3]

    assert jtu.tree_leaves(test.at[test == int].static()) == []
    assert jtu.tree_leaves(test.at[test == "a"].static()) == [2, 3]
    assert jtu.tree_leaves(test.at[(test == "a") | (test == "b")].static()) == [3]

    assert jtu.tree_leaves(test.at[test == {"a": 1}].static()) == [1, 2, 3]

    @pytc.treeclass
    class Test:
        a: jnp.ndarray = jnp.array([1, 2, 3])
        b: jnp.ndarray = jnp.array([4, 5, 6])

    test = Test()

    assert jtu.tree_leaves(test.at[...].static()) == []
    npt.assert_allclose(
        jtu.tree_leaves(test.at[test > 1].static())[0], jnp.array([1, 2, 3])
    )
    assert jtu.tree_leaves(test.at[test > 0].static()) == []

    @pytc.treeclass
    class Test:
        a: jnp.ndarray = static_field(default=jnp.array([1, 2, 3]))
        b: jnp.ndarray = static_field(default=jnp.array([4, 5, 6]))

    test = Test()

    assert jtu.tree_leaves(test) == []

    @pytc.treeclass
    class Test:
        a: jnp.ndarray = static_field(default=jnp.array([1, 2, 3]))
        b: jnp.ndarray = jnp.array([4, 5, 6])

    test = Test()
    npt.assert_allclose(jtu.tree_leaves(test)[0], jnp.array([4, 5, 6]))



from __future__ import annotations

import ctypes
from typing import Any

import jax.numpy as jnp
import requests

import pytreeclass
from pytreeclass.src.decorator_util import dispatch
from pytreeclass.src.tree_util import (
    _reduce_count_and_size,
    is_treeclass,
    is_treeclass_leaf,
    sequential_tree_shape_eval,
)
from pytreeclass.src.tree_viz_util import (
    _format_count,
    _format_node_diagram,
    _format_node_repr,
    _format_node_str,
    _format_size,
    _layer_box,
    _table,
    _vbox,
)

PyTree = Any


def tree_summary_md(tree: PyTree, array: jnp.ndarray | None = None) -> str:

    if array is not None:
        shape = sequential_tree_shape_eval(tree, array)
        indim_shape, outdim_shape = shape[:-1], shape[1:]

    def _cell(text):
        return f"<td align = 'center'> {text} </td>"

    def _leaf_info(tree_leaf: PyTree | Any) -> tuple[str, complex, complex]:
        """return (name, count, size) of a treeclass leaf / Any object"""

        @dispatch(argnum=0)
        def _info(leaf):
            """Any non-treeclass object"""
            count, size = _reduce_count_and_size(leaf)
            return (count, size)

        @_info.register(pytreeclass.src.tree_base.treeBase)
        def _(leaf):
            """treeclass leaf"""
            dynamic, static = leaf.__tree_fields__
            all_fields = {**dynamic, **static}
            count, size = _reduce_count_and_size(all_fields)
            return (count, size)

        return _info(tree_leaf)

    @dispatch(argnum=0)
    def recurse(tree, path=(), frozen_state=None):
        ...

    @recurse.register(pytreeclass.src.tree_base.treeBase)
    def _(tree, path=(), frozen_state=None):
        assert is_treeclass(tree)

        nonlocal FMT, COUNT, SIZE
        __all_fields__ = {
            **tree.__dataclass_fields__,
            **tree.__dict__.get("__treeclass_fields__", {}),
        }

        for i, fi in enumerate(__all_fields__.values()):

            cur_node = tree.__dict__[fi.name]

            if is_treeclass(cur_node) and not is_treeclass_leaf(cur_node):
                # Non leaf treeclass node
                recurse(
                    cur_node, path + (cur_node.__class__.__name__,), cur_node.frozen
                )

            elif is_treeclass_leaf(cur_node) or not is_treeclass(cur_node):
                # Leaf node (treeclass or non-treeclass)
                count, size = _leaf_info(cur_node)
                frozen_str = "<br>(frozen)" if frozen_state else ""
                name_str = f"{fi.name}{frozen_str}"
                type_str = "/".join(path + (cur_node.__class__.__name__,))
                count_str = _format_count(count, True)
                size_str = _format_size(size, True)
                config_str = (
                    "<br>".join(
                        [
                            f"{k}={_format_node_repr(v,0)}"
                            for k, v in cur_node.__tree_fields__[0].items()
                        ]
                    )
                    if is_treeclass(cur_node)
                    else f"{fi.name}={_format_node_repr(cur_node,0)}"
                )

                shape_str = (
                    f"{_format_node_repr(indim_shape[i],0)}\n{_format_node_repr(outdim_shape[i],0)}"
                    if array is not None
                    else ""
                )

                COUNT[1 if frozen_state else 0] += count
                SIZE[1 if frozen_state else 0] += size

                FMT += (
                    "<tr>"
                    + _cell(name_str)
                    + _cell(type_str)
                    + _cell(count_str)
                    + _cell(size_str)
                    + _cell(config_str)
                    + _cell(shape_str)
                    + "</tr>"
                )

    FMT = (
        "<table>\n"
        "<tr>\n"
        "<td align = 'center'> Name </td>\n"
        "<td align = 'center'> Type </td>\n"
        "<td align = 'center'> Param #</td>\n"
        "<td align = 'center'> Size </td>\n"
        "<td align = 'center'> Config </td>\n"
        "<td align = 'center'> Input/Output </td>\n"
        "</tr>\n"
    )

    COUNT = [0, 0]
    SIZE = [0, 0]

    recurse(tree, path=(), frozen_state=tree.frozen)

    FMT += "</table>"

    SUMMARY = (
        "<table>"
        f"<tr><td>Total #</td><td>{_format_count(sum(COUNT))}</td></tr>"
        f"<tr><td>Dynamic #</td><td>{_format_count(COUNT[0])}</td></tr>"
        f"<tr><td>Static/Frozen #</td><td>{_format_count(COUNT[1])}</td></tr>"
        f"<tr><td>Total size</td><td>{_format_size(sum(SIZE))}</td></tr>"
        f"<tr><td>Dynamic size</td><td>{_format_size(SIZE[0])}</td></tr>"
        f"<tr><td>Static/Frozen size</td><td>{_format_size(SIZE[1])}</td></tr>"
        "</table>"
    )

    return FMT + "\n\n#### Summary\n" + SUMMARY


def tree_summary(tree: PyTree, array: jnp.ndarray | None = None) -> str:
    assert is_treeclass(tree), "tree must be a treeclass object"

    if array is not None:
        shape = sequential_tree_shape_eval(tree, array)
        indim_shape, outdim_shape = shape[:-1], shape[1:]

    def _leaf_info(tree_leaf: PyTree | Any) -> tuple[str, complex, complex]:
        """return (name, count, size) of a treeclass leaf / Any object"""

        @dispatch(argnum=0)
        def _info(leaf):
            """Any object"""
            count, size = _reduce_count_and_size(leaf)
            return (count, size)

        @_info.register(pytreeclass.src.tree_base.treeBase)
        def _(leaf):
            """treeclass leaf"""
            dynamic, static = leaf.__tree_fields__
            all_fields = {**dynamic, **static}
            count, size = _reduce_count_and_size(all_fields)
            return (count, size)

        return _info(tree_leaf)

    @dispatch(argnum=0)
    def recurse(tree, path=(), frozen_state=None):
        ...

    @recurse.register(pytreeclass.src.tree_base.treeBase)
    def _(tree, path=(), frozen_state=None):
        _format_node = lambda node: _format_node_repr(node, depth=0).expandtabs(1)

        nonlocal ROWS, COUNT, SIZE

        __all_fields__ = {
            **tree.__dataclass_fields__,
            **tree.__dict__.get("__treeclass_fields__", {}),
        }

        for i, fi in enumerate(__all_fields__.values()):

            cur_node = tree.__dict__[fi.name]

            if is_treeclass(cur_node) and not is_treeclass_leaf(cur_node):
                # Non leaf treeclass node
                recurse(
                    cur_node, path + (cur_node.__class__.__name__,), cur_node.frozen
                )

            elif (
                is_treeclass_leaf(cur_node) or not is_treeclass(cur_node)
            ) and fi.repr:
                # Leaf node (treeclass or non-treeclass)
                count, size = _leaf_info(cur_node)
                frozen_str = "\n(frozen)" if frozen_state else ""
                name_str = f"{fi.name}{frozen_str}"
                type_str = "/".join(path + (cur_node.__class__.__name__,))
                count_str = _format_count(count, True)
                size_str = _format_size(size, True)
                config_str = (
                    "\n".join(
                        [
                            f"{k}={_format_node(v)}"
                            for k, v in cur_node.__tree_fields__[0].items()
                        ]
                    )
                    if is_treeclass(cur_node)
                    else f"{fi.name}={_format_node(cur_node)}"
                )

                shape_str = (
                    f"{_format_node(indim_shape[i])}\n{_format_node(outdim_shape[i])}"
                    if array is not None
                    else ""
                )

                COUNT[1 if frozen_state else 0] += count
                SIZE[1 if frozen_state else 0] += size

                ROWS.append(
                    [name_str, type_str, count_str, size_str, config_str, shape_str]
                )

    ROWS = [["Name", "Type ", "Param #", "Size ", "Config", "Input/Output"]]
    COUNT = [0, 0]
    SIZE = [0, 0]

    recurse(tree, path=(), frozen_state=tree.frozen)

    COLS = [list(c) for c in zip(*ROWS)]
    if array is None:
        COLS.pop()

    layer_table = _table(COLS)
    table_width = len(layer_table.split("\n")[0])

    param_summary = (
        f"Total # :\t\t{_format_count(sum(COUNT))}\n"
        f"Dynamic #:\t\t{_format_count(COUNT[0])}\n"
        f"Static/Frozen #:\t{_format_count(COUNT[1])}\n"
        f"{'-'*max([table_width,40])}\n"
        f"Total size :\t\t{_format_size(sum(SIZE))}\n"
        f"Dynamic size:\t\t{_format_size(SIZE[0])}\n"
        f"Static/Frozen size:\t{_format_size(SIZE[1])}\n"
        f"{'='*max([table_width,40])}"
    )

    return layer_table + "\n" + param_summary


def tree_box(tree, array=None):
    """
    === plot tree classes
    """

    def recurse(tree, parent_name):

        nonlocal shapes

        if is_treeclass_leaf(tree):
            frozen_stmt = "(Frozen)" if tree.frozen else ""
            box = _layer_box(
                f"{tree.__class__.__name__}[{parent_name}]{frozen_stmt}",
                _format_node_repr(shapes[0], 0) if array is not None else None,
                _format_node_repr(shapes[1], 0) if array is not None else None,
            )

            if shapes is not None:
                shapes.pop(0)
            return box

        else:
            level_nodes = []

            __all_fields__ = {
                **tree.__dataclass_fields__,
                **tree.__dict__.get("__treeclass_fields__", {}),
            }

            for fi in __all_fields__.values():
                cur_node = tree.__dict__[fi.name]

                if is_treeclass(cur_node):
                    level_nodes += [f"{recurse(cur_node,fi.name)}"]

                else:
                    level_nodes += [_vbox(f"{fi.name}={_format_node_repr(cur_node,0)}")]

            return _vbox(
                f"{tree.__class__.__name__}[{parent_name}]", "\n".join(level_nodes)
            )

    shapes = sequential_tree_shape_eval(tree, array) if array is not None else None
    return recurse(tree, "Parent")


def tree_diagram(tree):
    """
    === Explanation
        pretty print treeclass tree with tree structure diagram

    === Args
        tree : boolean to create tree-structure
    """
    assert is_treeclass(tree), "tree must be a treeclass object"

    @dispatch(argnum=1)
    def recurse_field(
        field_item, node_item, frozen_state, parent_level_count, node_index
    ):
        nonlocal FMT

        if field_item.repr:
            is_static_field = field_item.metadata.get("static", False)
            mark = "*" if is_static_field else ("#" if frozen_state else "─")
            is_last_field = node_index == 1

            FMT += "\n"
            FMT += "".join(
                [(("│" if lvl > 1 else "") + "\t") for lvl in parent_level_count]
            )

            FMT += f"└{mark}─ " if is_last_field else f"├{mark}─ "
            FMT += f"{field_item.name}"
            FMT += f"={_format_node_diagram(node_item)}"

        recurse(node_item, parent_level_count + [1], frozen_state)

    @recurse_field.register(pytreeclass.src.tree_base.treeBase)
    def _(field_item, node_item, frozen_state, parent_level_count, node_index):
        nonlocal FMT

        if field_item.repr:
            frozen_state = node_item.frozen
            is_static_field = field_item.metadata.get("static", False)
            mark = "*" if is_static_field else ("#" if frozen_state else "─")
            layer_class_name = node_item.__class__.__name__

            is_last_field = node_index == 1

            FMT += "\n" + "".join(
                [(("│" if lvl > 1 else "") + "\t") for lvl in parent_level_count]
            )

            FMT += f"└{mark}─ " if is_last_field else f"├{mark}─ "
            FMT += f"{field_item.name}"
            FMT += f"={layer_class_name}"

            recurse(node_item, parent_level_count + [node_index], frozen_state)

    @dispatch(argnum=0)
    def recurse(tree, parent_level_count, frozen_state):
        ...

    @recurse.register(pytreeclass.src.tree_base.treeBase)
    def _(tree, parent_level_count, frozen_state):
        nonlocal FMT

        __all_fields__ = {
            **tree.__dataclass_fields__,
            **tree.__dict__.get("__treeclass_fields__", {}),
        }
        leaves_count = len(__all_fields__)

        for i, fi in enumerate(__all_fields__.values()):

            cur_node = tree.__dict__[fi.name]

            recurse_field(
                fi, cur_node, frozen_state, parent_level_count, leaves_count - i
            )
        FMT += "\t"

    FMT = f"{(tree.__class__.__name__)}"

    recurse(tree, [1], tree.frozen)

    return FMT.expandtabs(4)


def tree_repr(tree, width: int = 40) -> str:
    """Prertty print `treeclass_leaves`

    Returns:
        str: indented tree leaves.
    """
    assert is_treeclass(tree), "tree must be a treeclass object"

    def format_width(string, width=width):
        """strip newline/tab characters if less than max width"""
        stripped_string = string.replace("\n", "").replace("\t", "")
        children_length = len(stripped_string)
        return string if children_length > width else stripped_string

    @dispatch(argnum=1)
    def recurse_field(field_item, node_item, depth, frozen_state, is_last_field):
        """format non-treeclass field"""
        nonlocal FMT

        if field_item.repr:
            is_static_field = field_item.metadata.get("static", False)
            mark = "*" if is_static_field else ("#" if frozen_state else "")

            FMT += "\n" + "\t" * depth
            FMT += f"{mark}{field_item.name}"
            FMT += f"={format_width(_format_node_repr(node_item,depth))}"
            FMT += "" if is_last_field else ","

        recurse(node_item, depth, frozen_state)

    @recurse_field.register(pytreeclass.src.tree_base.treeBase)
    def _(field_item, node_item, depth, frozen_state, is_last_field):
        """format treeclass field"""
        nonlocal FMT
        if field_item.repr:
            is_static_field = field_item.metadata.get("static", False)
            mark = "*" if is_static_field else ("#" if frozen_state else "")

            FMT += "\n" + "\t" * depth
            layer_class_name = f"{node_item.__class__.__name__}"

            FMT += f"{mark}{field_item.name}"
            FMT += f"={layer_class_name}" + "("

            start_cursor = len(FMT)  # capture children repr

            recurse(node_item, depth=depth + 1, frozen_state=node_item.frozen)

            FMT = FMT[:start_cursor] + format_width(FMT[start_cursor:]) + ")"
            FMT += "" if is_last_field else ","

    @dispatch(argnum=0)
    def recurse(tree, depth, frozen_state):
        ...

    @recurse.register(pytreeclass.src.tree_base.treeBase)
    def _(tree, depth, frozen_state):
        nonlocal FMT
        is_treeclass(tree)

        __all_fields__ = {
            **tree.__dataclass_fields__,
            **tree.__dict__.get("__treeclass_fields__", {}),
        }

        leaves_count = len(__all_fields__)
        for i, fi in enumerate(__all_fields__.values()):

            # retrieve node item
            cur_node = tree.__dict__[fi.name]

            recurse_field(
                fi,
                cur_node,
                depth,
                frozen_state,
                True if i == (leaves_count - 1) else False,
            )

    FMT = ""
    recurse(tree, depth=1, frozen_state=tree.frozen)
    FMT = f"{(tree.__class__.__name__)}({format_width(FMT,width)})"

    return FMT.expandtabs(2)


def tree_str(tree, width: int = 40) -> str:
    """Prertty print `treeclass_leaves`

    Returns:
        str: indented tree leaves.
    """
    assert is_treeclass(tree), "tree must be a treeclass object"

    def format_width(string, width=width):
        """strip newline/tab characters if less than max width"""
        stripped_string = string.replace("\n", "").replace("\t", "")
        children_length = len(stripped_string)
        return string if children_length > width else stripped_string

    @dispatch(argnum=1)
    def recurse_field(field_item, node_item, depth, frozen_state, is_last_field):
        """format non-treeclass field"""
        nonlocal FMT

        if field_item.repr:
            is_static_field = field_item.metadata.get("static", False)
            mark = "*" if is_static_field else ("#" if frozen_state else "")

            FMT += "\n" + "\t" * depth
            FMT += f"{mark}{field_item.name}"
            FMT += f"={format_width(_format_node_str(node_item,depth))}"
            FMT += "" if is_last_field else ","

        recurse(node_item, depth, frozen_state)

    @recurse_field.register(pytreeclass.src.tree_base.treeBase)
    def _(field_item, node_item, depth, frozen_state, is_last_field):
        """format treeclass field"""
        nonlocal FMT

        if field_item.repr:
            is_static_field = field_item.metadata.get("static", False)
            mark = "*" if is_static_field else ("#" if frozen_state else "")

            FMT += "\n" + "\t" * depth
            layer_class_name = f"{node_item.__class__.__name__}"

            FMT += f"{mark}{field_item.name}"
            FMT += f"={layer_class_name}" + "("
            start_cursor = len(FMT)  # capture children repr

            recurse(node_item, depth=depth + 1, frozen_state=node_item.frozen)

            FMT = FMT[:start_cursor] + format_width(FMT[start_cursor:]) + ")"
            FMT += "" if is_last_field else ","

    @dispatch(argnum=0)
    def recurse(tree, depth, frozen_state):
        ...

    @recurse.register(pytreeclass.src.tree_base.treeBase)
    def _(tree, depth, frozen_state):
        nonlocal FMT
        assert is_treeclass(tree)

        __all_fields__ = {
            **tree.__dataclass_fields__,
            **tree.__dict__.get("__treeclass_fields__", {}),
        }

        leaves_count = len(__all_fields__)
        for i, fi in enumerate(__all_fields__.values()):

            # retrieve node item
            cur_node = tree.__dict__[fi.name]

            recurse_field(
                fi,
                cur_node,
                depth,
                frozen_state,
                True if i == (leaves_count - 1) else False,
            )

    FMT = ""
    recurse(tree, depth=1, frozen_state=tree.frozen)
    FMT = f"{(tree.__class__.__name__)}({format_width(FMT,width)})"

    return FMT.expandtabs(2)


def _tree_mermaid(tree):
    assert is_treeclass(tree), "tree must be a treeclass object"

    def node_id(input):
        """hash a node by its location in a tree"""
        return ctypes.c_size_t(hash(input)).value

    @dispatch(argnum=1)
    def recurse_field(field_item, node_item, depth, prev_id, order, frozen_state):
        nonlocal FMT

        if field_item.repr:
            # create node id from depth, order, and previous id
            cur_id = node_id((depth, order, prev_id))
            mark = (
                "--x"
                if field_item.metadata.get("static", False)
                else ("-.-" if frozen_state else "---")
            )
            FMT += f'\n\tid{prev_id} {mark} id{cur_id}["{field_item.name}\\n{_format_node_diagram(node_item)}"]'
            prev_id = cur_id

        recurse(node_item, depth, prev_id, frozen_state)

    @recurse_field.register(pytreeclass.src.tree_base.treeBase)
    def _(field_item, node_item, depth, prev_id, order, frozen_state):
        nonlocal FMT

        if field_item.repr:
            layer_class_name = node_item.__class__.__name__
            cur_id = node_id((depth, order, prev_id))
            FMT += f"\n\tid{prev_id} --> id{cur_id}({field_item.name}\\n{layer_class_name})"
            recurse(node_item, depth + 1, cur_id, node_item.frozen)

    @dispatch(argnum=0)
    def recurse(tree, depth, prev_id, frozen_state):
        ...

    @recurse.register(pytreeclass.src.tree_base.treeBase)
    def _(tree, depth, prev_id, frozen_state):
        nonlocal FMT
        assert is_treeclass(tree)

        __all_fields__ = {
            **tree.__dataclass_fields__,
            **tree.__dict__.get("__treeclass_fields__", {}),
        }

        for i, fi in enumerate(__all_fields__.values()):

            # retrieve node item
            cur_node = tree.__dict__[fi.name]

            recurse_field(
                fi,
                cur_node,
                depth,
                prev_id,
                i,
                frozen_state,
            )

    cur_id = node_id((0, 0, -1, 0))
    FMT = f"flowchart LR\n\tid{cur_id}[{tree.__class__.__name__}]"
    recurse(tree, 1, cur_id, tree.frozen)
    return FMT.expandtabs(4)


def _generate_mermaid_link(mermaid_string: str) -> str:
    """generate a one-time link mermaid diagram"""
    url_val = "https://pytreeclass.herokuapp.com/generateTemp"
    request = requests.post(url_val, json={"description": mermaid_string})
    generated_id = request.json()["id"]
    generated_html = f"https://pytreeclass.herokuapp.com/temp/?id={generated_id}"
    return f"Open URL in browser: {generated_html}"


def tree_mermaid(tree, link=False):
    mermaid_string = _tree_mermaid(tree)
    return _generate_mermaid_link(mermaid_string) if link else mermaid_string


def save_viz(tree, filename, method="tree_mermaid_md"):

    if method == "tree_mermaid_md":
        FMT = "```mermaid\n" + tree_mermaid(tree) + "\n```"

        with open(f"{filename}.md", "w") as f:
            f.write(FMT)

    elif method == "tree_mermaid_html":
        FMT = "<html><body><script src='https://cdn.jsdelivr.net/npm/mermaid/dist/mermaid.min.js'></script>"
        FMT += "<script>mermaid.initialize({ startOnLoad: true });</script><div class='mermaid'>"
        FMT += tree_mermaid(tree)
        FMT += "</div></body></html>"

        with open(f"{filename}.html", "w") as f:
            f.write(FMT)

    elif method == "tree_diagram":
        with open(f"{filename}.txt", "w") as f:
            f.write(tree_diagram(tree))

    elif method == "tree_box":
        with open(f"{filename}.txt", "w") as f:
            f.write(tree_box(tree))

    elif method == "summary":
        with open(f"{filename}.txt", "w") as f:
            f.write(tree_summary(tree))

    elif method == "summary_md":
        with open(f"{filename}.md", "w") as f:
            f.write(tree_summary_md(tree))



from __future__ import annotations

from dataclasses import MISSING, field
from typing import Any, Sequence

import jax.numpy as jnp
import jax.tree_util as jtu

from pytreeclass.src.tree_util import (
    _freeze_nodes,
    _unfreeze_nodes,
    is_treeclass,
    is_treeclass_leaf,
    static_value,
)
from pytreeclass.src.tree_viz import (
    tree_box,
    tree_diagram,
    tree_repr,
    tree_str,
    tree_summary,
)

PyTree = Any


class fieldDict(dict):
    # dict will throw
    # `ValueError: The truth value of an array with more than one element is ambiguous. Use a.any() or a.all()``
    # while this implementation will not.
    # relevant issue : https://github.com/google/jax/issues/11089
    def __eq__(self, other):
        return self.__dict__ == other.__dict__


class treeBase:
    def freeze(self) -> PyTree:
        """Freeze treeclass.

        Returns:
            New frozen instance.

        Example:

        >>> model = model.freeze()
        >>> assert model.frozen == True
        """
        return _freeze_nodes(jtu.tree_unflatten(*jtu.tree_flatten(self)[::-1]))

    def unfreeze(self) -> PyTree:
        """Unfreeze treeclass.

        Returns:
            New unfrozen instance.

        Example :

        >>> model = model.unfreeze()
        >>> assert model.frozen == False
        """
        return _unfreeze_nodes(jtu.tree_unflatten(*jtu.tree_flatten(self)[::-1]))

    @property
    def frozen(self) -> bool:
        """Show treeclass frozen status.

        Returns:
            Frozen state boolean.
        """
        return True if hasattr(self, "__frozen_fields__") else False

    def tree_flatten(self):
        """Flatten rule for `jax.tree_flatten`

        Returns:
            Tuple of dynamic values and (dynamic keys,static dict)
        """
        # we need to transfer the state for the next instance through static
        # we also need to retrieve it for the current instance

        dynamic, static = self.__tree_fields__

        if self.frozen:
            return (), ((), {"__frozen_fields__": (dynamic, static)})

        else:
            return dynamic.values(), (dynamic.keys(), static)

    @classmethod
    def tree_unflatten(cls, treedef, children):
        """Unflatten rule for `jax.tree_unflatten`

        Args:
            treedef:
                Pytree definition
                includes Dynamic nodes keys , static dictionary and frozen state
            children:
                Dynamic nodes values

        Returns:
            New class instance
        """

        # new_cls = type(cls.__name__, cls.__bases__[1:], dict(cls.__dict__))
        new_cls = cls.__new__(cls)

        tree_fields = treedef[1].get("__frozen_fields__", None)

        if tree_fields is not None:
            object.__setattr__(new_cls, "__frozen_fields__", tree_fields)
            dynamic, static = tree_fields
            attrs = {**dynamic, **static}

        else:

            dynamic_vals, dynamic_keys = children, treedef[0]
            static_keys, static_vals = treedef[1].keys(), treedef[1].values()

            attrs = dict(
                zip(
                    (*dynamic_keys, *static_keys),
                    (*dynamic_vals, *static_vals),
                )
            )

        for k, v in attrs.items():
            object.__setattr__(new_cls, k, v)
        return new_cls

    @property
    def treeclass_leaves(self) -> Sequence[PyTree | Any, ...]:
        """Tree leaves of treeclass

        Example:

            @pytc.treeclass
            class T0:
                a : int = 1
                b : int = 2

            @pytc.treeclass
            class T1:
                c : T0 = T0()
                d : int = 3


            @pytc.treeclass
            class T2 :
                e : T1 = T1()
                f : T0 = T0()
                g : int = 4

            >>> t = T2()

            >>> print(t.tree_diagram())
            T2
                ├── e=T1
                │   ├── c=T0
                │   │   ├── a=1
                │   │   └── b=2
                │   └── d=3
                ├── f=T0
                │   ├── a=1
                │   └── b=2
                └── g=4

            >>> print(t.treeclass_leaves)
            [T0(a=1,b=2), 3, T0(a=1,b=2), 4]
        """
        return jtu.tree_leaves(self, is_treeclass_leaf)

    def __hash__(self):
        return hash(tuple(*jtu.tree_flatten(self)))

    def asdict(self) -> dict[str, Any]:
        """Dictionary representation of dataclass_fields"""
        dynamic, static = self.__tree_fields__
        static.pop("__treeclass_fields__", None)
        return {
            **dynamic,
            **jtu.tree_map(
                lambda x: x.value if isinstance(x, static_value) else x, dict(static)
            ),
        }

    def register_node(
        self, node: Any, *, name: str, static: bool = False, repr: bool = True
    ) -> Any:
        """Add item to dataclass fields to bee seen by jax computations"""
        __all_fields__ = {
            **self.__dataclass_fields__,
            **self.__dict__.get("__treeclass_fields__", {}),
        }

        if name not in __all_fields__:
            # create field
            field_value = field(repr=repr, metadata={"static": static})

            object.__setattr__(field_value, "name", name)
            object.__setattr__(field_value, "type", type(node))

            # register it to class
            __treeclass_fields__ = self.__dict__.get("__treeclass_fields__", {})
            __treeclass_fields__[name] = field_value
            object.__setattr__(self, "__treeclass_fields__", __treeclass_fields__)
            object.__setattr__(self, name, node)

        return self.__dict__[name]

    def __repr__(self):
        return tree_repr(self)

    def __str__(self):
        return tree_str(self)

    def summary(self, array: jnp.ndarray = None) -> str:
        return tree_summary(self, array)

    def tree_diagram(self) -> str:
        return tree_diagram(self)

    def tree_box(self, array: jnp.ndarray = None) -> str:
        return tree_box(self, array)

    def __generate_tree_fields__(self) -> tuple[dict[str, Any], dict[str, Any]]:
        dynamic, static = fieldDict(), fieldDict()
        # register other variables defined in other context
        # if their value is an instance of treeclass
        # to avoid redefining them as dataclass fields.

        # register *all* dataclass fields
        __all_fields__ = {
            **self.__dataclass_fields__,
            **self.__dict__.get("__treeclass_fields__", {}),
        }

        for fi in __all_fields__.values():
            # field value is defined in class dict
            if fi.name in self.__dict__:
                value = self.__dict__[fi.name]

            # field value is defined in field default
            elif fi.default is not MISSING:
                self.__dict__[fi.name] = fi.default
                value = fi.default

            else:
                # the user did not declare a variable defined in field
                raise ValueError(f"field={fi.name} is not declared.")

            excluded_by_meta = fi.metadata.get("static", False)
            excluded_by_type = isinstance(value, static_value)

            if excluded_by_type or excluded_by_meta:
                static[fi.name] = value

            else:
                dynamic[fi.name] = value

        static["__treeclass_fields__"] = self.__dict__.get("__treeclass_fields__", {})

        return (dynamic, static)


class explicitTreeBase:
    """ "Register  dataclass fields only"""

    @property
    def __tree_fields__(self):
        """Computes the dynamic and static fields.

        Returns:
            Pair of dynamic and static dictionaries.
        """
        if self.__dict__.get("__frozen_fields__", None) is not None:
            return self.__frozen_fields__
        else:
            return self.__generate_tree_fields__()

    def __setattr__(self, name: str, value: Any) -> None:
        if self.frozen:
            raise ValueError("Cannot set a value to a frozen treeclass.")
        object.__setattr__(self, name, value)


class implicitTreeBase:
    """Register dataclass fields and treeclass instance variables"""

    def __register_treeclass_instance_variables__(self) -> None:

        __treeclass_fields__ = {}

        for var_name, var_value in self.__dict__.items():
            # check if a variable in self.__dict__ is treeclass
            # that is not defined in fields
            if (
                isinstance(var_name, str)
                and is_treeclass(var_value)
                and var_name not in self.__dataclass_fields__
            ):

                field_value = field()

                object.__setattr__(field_value, "name", var_name)
                object.__setattr__(field_value, "type", type(var_value))

                # register it to class
                __treeclass_fields__.update({var_name: field_value})

        if len(__treeclass_fields__) > 0:
            object.__setattr__(self, "__treeclass_fields__", __treeclass_fields__)

    def __setattr__(self, name: str, value: Any) -> None:
        if self.frozen:
            raise ValueError("Cannot set a value to a frozen treeclass.")
        object.__setattr__(self, name, value)
        self.__register_treeclass_instance_variables__()

    @property
    def __tree_fields__(self):
        """Computes the dynamic and static fields.

        Returns:
            Pair of dynamic and static dictionaries.
        """
        if self.__dict__.get("__frozen_fields__", None) is not None:
            return self.__frozen_fields__

        else:
            return self.__generate_tree_fields__()



from __future__ import annotations

import functools as ft
from collections.abc import Callable
from typing import Any

import jax
import jax.numpy as jnp
import jax.tree_util as jtu

from pytreeclass.src.decorator_util import dispatch
from pytreeclass.src.tree_util import is_treeclass_leaf_bool, static_value

""" Getter """


def _at_get(tree, where, **kwargs):
    @dispatch(argnum=0)
    def _node_get(lhs: Any, where: Any, **kwargs):
        """Get pytree node  value

        Args:
            lhs (Any): Node value.
            where (Any): Conditional

        Raises:
            NotImplementedError:
        """
        # not jittable as size can changes
        # does not change pytreestructure ,
        raise NotImplementedError(f"Get node type ={type(lhs)} is not implemented.")

    @_node_get.register(jax.interpreters.partial_eval.DynamicJaxprTracer)
    @_node_get.register(jnp.ndarray)
    def _(lhs, where, array_as_leaves: bool = True):
        return (
            (lhs[jnp.where(where)])
            if array_as_leaves
            else (lhs if jnp.all(where) else None)
        )

    @_node_get.register(int)
    @_node_get.register(float)
    @_node_get.register(complex)
    @_node_get.register(tuple)
    @_node_get.register(list)
    @_node_get.register(str)
    def _(lhs, where, **kwargs):
        # set None to non-chosen non-array values
        return lhs if where else None

    @dispatch(argnum=1)
    def __at_get(tree, where, **kwargs):
        raise NotImplementedError(f"Get where type = {type(where)} is not implemented.")

    @__at_get.register(type(tree))
    def _(tree, where, is_leaf=None, **kwargs):
        lhs_leaves, lhs_treedef = jtu.tree_flatten(tree, is_leaf=is_leaf)
        where_leaves, where_treedef = jtu.tree_flatten(where, is_leaf=is_leaf)
        lhs_leaves = [
            _node_get(lhs_leaf, where_leaf, **kwargs)
            for lhs_leaf, where_leaf in zip(lhs_leaves, where_leaves)
        ]

        return jtu.tree_unflatten(lhs_treedef, lhs_leaves)

    return __at_get(tree, where, **kwargs)


""" Setter """


def _at_set(tree, where, set_value, **kwargs):
    @dispatch(argnum=0)
    def _node_set(lhs: Any, where: bool, set_value, **kwargs):
        """Set pytree node value.

        Args:
            lhs: Node value.
            where: Conditional.
            set_value: Set value of shape 1.

        Returns:
            Modified node value.
        """
        raise NotImplementedError(f"Set node type = {type(lhs)} is unknown.")

    @_node_set.register(jax.interpreters.partial_eval.DynamicJaxprTracer)
    @_node_set.register(jnp.ndarray)
    @dispatch(argnum=2)
    def _array_node(lhs, where, set_value):
        """Multi dispatched on lhs and where type"""
        # lhs is numeric node
        # set_value in not acceptable set_value type for numeric node
        # thus array_as_leaves is not an optional keyword argument
        # An example is setting an array to None
        # then the entire array is set to None if all array elements are
        # satisfied by the where condition
        return set_value if jnp.all(where) else lhs

    @_array_node.register(bool)
    def _(lhs, where, set_value):
        # in python isinstance(True/False,int) is True
        # without this dispatch, it will be handled with the int dispatch
        return set_value if jnp.all(where) else lhs

    @_array_node.register(int)
    @_array_node.register(float)
    @_array_node.register(complex)
    @_array_node.register(jnp.ndarray)
    def _(lhs, where, set_value, array_as_leaves: bool = True):
        # lhs is numeric node
        # set_value in acceptable set_value type for a numeric node
        # For some reason python isinstance(True,int) is True ?
        return (
            jnp.where(where, set_value, lhs)
            if array_as_leaves
            else (set_value if jnp.all(where) else lhs)
        )

    @_node_set.register(int)
    @_node_set.register(float)
    @_node_set.register(complex)
    @_node_set.register(tuple)
    @_node_set.register(list)
    @_node_set.register(str)
    @_node_set.register(type(None))
    def _(lhs, where, set_value, **kwargs):
        # where == None can be obtained by
        # is_leaf = lambda x : x is None
        return set_value if (where in [True, None]) else lhs

    @dispatch(argnum=1)
    def __at_set(tree, where, set_value, is_leaf=None, **kwargs):
        raise NotImplementedError(f"Set where type = {type(where)} is not implemented.")

    @__at_set.register(type(tree))
    def _(tree, where, set_value, is_leaf=None, **kwargs):
        lhs_leaves, lhs_treedef = jtu.tree_flatten(tree, is_leaf=is_leaf)
        where_leaves, rhs_treedef = jtu.tree_flatten(where, is_leaf=is_leaf)
        lhs_leaves = [
            _node_set(lhs_leaf, where_leaf, set_value, **kwargs)
            for lhs_leaf, where_leaf in zip(lhs_leaves, where_leaves)
        ]

        return jtu.tree_unflatten(lhs_treedef, lhs_leaves)

    return __at_set(tree, where, set_value, **kwargs)


""" Apply """


def _at_apply(tree, where, func, **kwargs):
    @dispatch(argnum=0)
    def _node_apply(lhs: Any, where: bool, func: Callable[[Any], Any], **kwargs):
        """Set pytree node

        Args:
            lhs (Any): Node value.
            where (bool): Conditional
            func (Callable[[Any], Any]): Callable

        Raises:
            NotImplementedError:
        """
        raise NotImplementedError(f"Apply node type= {type(lhs)} is not implemented.")

    @_node_apply.register(jax.interpreters.partial_eval.DynamicJaxprTracer)
    @_node_apply.register(jnp.ndarray)
    def _(lhs, where, func, array_as_leaves: bool = True):
        return (
            jnp.where(where, func(lhs), lhs)
            if array_as_leaves
            else (func(lhs) if jnp.all(where) else lhs)
        )

    @_node_apply.register(int)
    @_node_apply.register(float)
    @_node_apply.register(complex)
    @_node_apply.register(tuple)
    @_node_apply.register(list)
    @_node_apply.register(str)
    @_node_apply.register(type(None))
    def _(lhs, where, func, **kwargs):
        return func(lhs) if (where in [True, None]) else lhs

    @dispatch(argnum=1)
    def __at_apply(tree, where, func, **kwargs):
        raise NotImplementedError(
            f"Apply where type = {type(where)} is not implemented."
        )

    @__at_apply.register(type(tree))
    def _(tree, where, func, is_leaf=None, **kwargs):

        lhs_leaves, lhs_treedef = jtu.tree_flatten(tree, is_leaf=is_leaf)
        where_leaves, rhs_treedef = jtu.tree_flatten(where, is_leaf=is_leaf)
        lhs_leaves = [
            _node_apply(lhs_leaf, where_leaf, func, **kwargs)
            for lhs_leaf, where_leaf in zip(lhs_leaves, where_leaves)
        ]

        return jtu.tree_unflatten(lhs_treedef, lhs_leaves)

    return __at_apply(tree, where, func, **kwargs)


""" Reduce """


def _at_reduce(tree, where, func, **kwargs):
    @dispatch(argnum=1)
    def __at_reduce(tree, where, func, **kwargs):
        raise NotImplementedError(
            f"Reduce where type = {type(where)} is not implemented."
        )

    @__at_reduce.register(type(tree))
    def _(tree, where, func, initializer=0, **kwargs):
        return jtu.tree_reduce(func, tree.at[where].get(), initializer)

    return __at_reduce(tree, where, func, **kwargs)


""" Static"""


def _at_static(tree, where, **kwargs):
    def __at_static(tree, where, **kwargs):
        return tree.at[where].apply(static_value, array_as_leaves=False)

    return __at_static(tree, where, **kwargs)


class treeIndexerMethods:
    def add(getter_setter_self, set_value):
        return getter_setter_self.apply(lambda x: x + set_value)

    def multiply(getter_setter_self, set_value):
        return getter_setter_self.apply(lambda x: x * set_value)

    def divide(getter_setter_self, set_value):
        return getter_setter_self.apply(lambda x: x / set_value)

    def power(getter_setter_self, set_value):
        return getter_setter_self.apply(lambda x: x**set_value)

    def min(getter_setter_self, set_value):
        return getter_setter_self.apply(lambda x: jnp.minimum(x, set_value))

    def max(getter_setter_self, set_value):
        return getter_setter_self.apply(lambda x: jnp.maximum(x, set_value))

    def reduce_sum(getter_setter_self):
        return getter_setter_self.reduce(lambda x, y: x + jnp.sum(y))

    def reduce_product(getter_setter_self):
        return getter_setter_self.reduce(lambda x, y: x * jnp.prod(y), initializer=1)

    def reduce_max(getter_setter_self):
        return getter_setter_self.reduce(
            lambda x, y: jnp.maximum(x, jnp.max(y)), initializer=-jnp.inf
        )

    def reduce_min(getter_setter_self):
        return getter_setter_self.reduce(
            lambda x, y: jnp.minimum(x, jnp.min(y)), initializer=+jnp.inf
        )


class treeIndexer:
    @property
    def at(self):
        class indexer:
            @dispatch(argnum=1)
            def __getitem__(inner_self, *args):
                raise NotImplementedError(
                    f"Indexing with type{tuple(type(arg) for arg in args)} is not implemented."
                )

            @__getitem__.register(type(self))
            def _(inner_self, arg):
                """indexing by boolean pytree"""

                assert all(
                    is_treeclass_leaf_bool(leaf) for leaf in jtu.tree_leaves(arg)
                ), f"All tree leaves must be boolean.Found {jtu.tree_leaves(arg)}"

                class getterSetterIndexer(treeIndexerMethods):
                    def get(getter_setter_self, **kwargs):
                        return ft.partial(_at_get, where=arg)(tree=self, **kwargs)

                    def set(getter_setter_self, set_value, **kwargs):
                        if self.frozen:
                            raise ValueError("Cannot set to a frozen treeclass.")
                        return ft.partial(_at_set, where=arg)(
                            tree=self, set_value=set_value, **kwargs
                        )

                    def apply(getter_setter_self, func, **kwargs):
                        if self.frozen:
                            raise ValueError("Cannot apply to a frozen treeclass.")
                        return ft.partial(_at_apply, where=arg)(
                            tree=self, func=func, **kwargs
                        )

                    def reduce(getter_setter_self, func, **kwargs):
                        if self.frozen:
                            raise ValueError("Cannot reduce to a frozen treeclass.")
                        return ft.partial(_at_reduce, where=arg)(
                            tree=self, func=func, **kwargs
                        )

                    def static(getter_setter_self, **kwargs):
                        if self.frozen:
                            raise ValueError(
                                "Cannot apply static to a frozen treeclass."
                            )
                        return ft.partial(_at_static, where=arg)(tree=self, **kwargs)

                return getterSetterIndexer()

            @__getitem__.register(type(Ellipsis))
            def _(inner_self, arg):
                """Ellipsis as an alias for all elements"""
                return self.at.__getitem__(self == self)

        return indexer()



from __future__ import annotations

import functools
import inspect

""" Porting utils """


class cached_property:
    def __init__(self, func):
        self.name = func.__name__
        self.func = func

    def __get__(self, instance, owner):
        attr = self.func(instance)
        object.__setattr__(instance, self.name, attr)
        return attr


def dispatch(*args, **kwargs):
    """functools.singledispatch with option of choosing the position of the dispatched argument
    or keyword argument.

    For multiple dispatch use the following pattern

    >>> @dispatch(argnum=0)
    ... def a(x,y): ...

    >>> @a.register(int)
    ... @dispatch(argnum=1)
    ... def b(x,y) : ...

    >>> @b.register(int)
    ... def _(x,y):
    ...     return "int,int"
    """

    def dispatch_wrapper(func, argnum: int | str = 0):
        """singledispatch by arg position/kw arg name"""
        dispatcher = functools.singledispatch(func)

        def wrapper(*ar, **kw):
            if isinstance(argnum, int):
                # based on https://stackoverflow.com/a/24602374/10879163
                return dispatcher.dispatch(ar[argnum].__class__)(*ar, **kw)

            elif isinstance(argnum, str):
                # dispatch by keyword argument
                return dispatcher.dispatch(kw[argnum].__class__)(*ar, **kw)

            else:
                raise ValueError("argnum must be int or str")

        wrapper.register = dispatcher.register
        functools.update_wrapper(wrapper, func)
        return wrapper

    if len(args) == 1 and inspect.isfunction(args[0]):
        # @dispatch
        # def f(..):
        return dispatch_wrapper(args[0], 0)

    elif len(args) == 0 and len(kwargs) > 0:
        # @dispatch(argnum=...)
        # def f(..):
        argnum = kwargs["argnum"] if "argnum" in kwargs else 0
        return functools.partial(dispatch_wrapper, argnum=argnum)



from __future__ import annotations

import jax

from pytreeclass.src import tree_viz_util


def test__vbox():

    assert (
        tree_viz_util._vbox("a", " a", "a ")
        == "┌──┐\n│a │\n├──┤\n│ a│\n├──┤\n│a │\n└──┘"
    )


def test__hbox():
    assert tree_viz_util._hbox("a", "b", "c") == "┌─┬─┬─┐\n│a│b│c│\n└─┴─┴─┘\n"


def test_func_repr():
    def example(a: int, b=1, *c, d, e=2, **f) -> str: ...

    assert tree_viz_util._func_repr(example) == "example(a,b,*c,d,e,**f)"
    assert tree_viz_util._func_repr(lambda x: x) == "Lambda(x)"
    assert tree_viz_util._func_repr(jax.nn.relu) == "relu(*args,**kwargs)"
    assert (
        tree_viz_util._func_repr(jax.nn.initializers.he_normal)
        == "he_normal(in_axis,out_axis,batch_axis,dtype)"
    )



from __future__ import annotations

import functools
import operator as op

import jax
import jax.numpy as jnp
import jax.tree_util as jtu

from pytreeclass.src.decorator_util import dispatch
from pytreeclass.src.tree_util import _dispatched_tree_map, is_excluded, is_treeclass


def _append_math_op(func):
    """binary and unary magic operations"""

    @functools.wraps(func)
    def wrapper(self, rhs=None):
        return _dispatched_tree_map(func, self, rhs)

    return wrapper


def _append_math_eq_ne(func):
    """Append eq/ne operations"""
    assert func in [op.eq, op.ne], f"func={func} is not implemented."

    def node_not(node):
        if isinstance(node, jnp.ndarray):
            return jnp.logical_not(node)
        else:
            return not node

    def set_true(node, array_as_leaves: bool = True):
        if isinstance(node, jnp.ndarray):
            return jnp.ones_like(node).astype(jnp.bool_) if array_as_leaves else True
        else:
            return True

    def set_false(node, array_as_leaves: bool = True):
        if isinstance(node, jnp.ndarray):
            return jnp.zeros_like(node).astype(jnp.bool_) if array_as_leaves else False
        else:
            return False

    @functools.wraps(func)
    def wrapper(self, rhs):
        @dispatch(argnum=1)
        def inner_wrapper(tree, where, **kwargs):
            raise NotImplementedError(f"rhs of type {type(rhs)} is not implemented.")

        @inner_wrapper.register(int)
        @inner_wrapper.register(float)
        @inner_wrapper.register(complex)
        @inner_wrapper.register(bool)
        @inner_wrapper.register(type(self))
        @inner_wrapper.register(jax.interpreters.partial_eval.DynamicJaxprTracer)
        def _(self, rhs):
            return _dispatched_tree_map(func, self, rhs)

        @inner_wrapper.register(str)
        def _(tree, where, **kwargs):
            """Filter by field name"""
            tree_copy = jtu.tree_unflatten(*jtu.tree_flatten(tree)[::-1])

            def recurse(tree, where, **kwargs):
                __all_fields__ = {
                    **tree.__dataclass_fields__,
                    **tree.__dict__.get("__treeclass_fields__", {}),
                }
                for i, fld in enumerate(__all_fields__.values()):

                    cur_node = tree.__dict__[fld.name]
                    if not is_excluded(fld, cur_node) and is_treeclass(cur_node):
                        if fld.name == where:
                            # broadcast True to all subtrees
                            tree.__dict__[fld.name] = jtu.tree_map(set_true, cur_node)
                        else:
                            recurse(cur_node, where, **kwargs)
                    else:
                        tree.__dict__[fld.name] = (
                            set_true(cur_node)
                            if ((fld.name == where))
                            else set_false(cur_node)
                        )
                return tree

            return recurse(tree_copy, where, **kwargs)

        @inner_wrapper.register(type)
        def _(tree, where, **kwargs):
            """Filter by type"""
            tree_copy = jtu.tree_unflatten(*jtu.tree_flatten(tree)[::-1])

            def recurse(tree, where, **kwargs):

                __all_fields__ = {
                    **tree.__dataclass_fields__,
                    **tree.__dict__.get("__treeclass_fields__", {}),
                }
                for i, fld in enumerate(__all_fields__.values()):

                    cur_node = tree.__dict__[fld.name]

                    if not is_excluded(fld, cur_node) and is_treeclass(cur_node):
                        if isinstance(cur_node, where):
                            tree.__dict__[fld.name] = jtu.tree_map(set_true, cur_node)
                        else:
                            recurse(cur_node, where, **kwargs)
                    else:
                        tree.__dict__[fld.name] = (
                            set_true(cur_node)
                            if (isinstance(cur_node, where))
                            else set_false(cur_node)
                        )

                return tree

            return recurse(tree_copy, where, **kwargs)

        @inner_wrapper.register(dict)
        def _(tree, where, **kwargs):
            """Filter by metadata"""
            tree_copy = jtu.tree_unflatten(*jtu.tree_flatten(tree)[::-1])

            def in_metadata(fld):
                kws, vals = zip(*where.items())
                return all(
                    fld.metadata.get(kw, False) and (fld.metadata[kw] == val)
                    for kw, val in zip(kws, vals)
                )

            def recurse(tree, where, **kwargs):

                __all_fields__ = {
                    **tree.__dataclass_fields__,
                    **tree.__dict__.get("__treeclass_fields__", {}),
                }
                for i, fld in enumerate(__all_fields__.values()):

                    cur_node = tree.__dict__[fld.name]

                    if not is_excluded(fld, cur_node) and is_treeclass(cur_node):
                        if in_metadata(fld):
                            tree.__dict__[fld.name] = jtu.tree_map(set_true, cur_node)
                        else:
                            recurse(cur_node, where, **kwargs)
                    else:
                        tree.__dict__[fld.name] = (
                            set_true(cur_node)
                            if in_metadata(fld)
                            else set_false(cur_node)
                        )

                return tree

            return recurse(tree_copy, where, **kwargs)

        return (
            jtu.tree_map(node_not, inner_wrapper(self, rhs))
            if func == op.ne
            else inner_wrapper(self, rhs)
        )

    return wrapper


class treeOpBase:

    __abs__ = _append_math_op(op.abs)
    __add__ = _append_math_op(op.add)
    __radd__ = _append_math_op(op.add)
    __and__ = _append_math_op(op.and_)
    __rand__ = _append_math_op(op.and_)
    __eq__ = _append_math_eq_ne(op.eq)
    __floordiv__ = _append_math_op(op.floordiv)
    __ge__ = _append_math_op(op.ge)
    __gt__ = _append_math_op(op.gt)
    __inv__ = _append_math_op(op.inv)
    __invert__ = _append_math_op(op.invert)
    __le__ = _append_math_op(op.le)
    __lshift__ = _append_math_op(op.lshift)
    __lt__ = _append_math_op(op.lt)
    __matmul__ = _append_math_op(op.matmul)
    __mod__ = _append_math_op(op.mod)
    __mul__ = _append_math_op(op.mul)
    __rmul__ = _append_math_op(op.mul)
    __ne__ = _append_math_eq_ne(op.ne)
    __neg__ = _append_math_op(op.neg)
    __not__ = _append_math_op(op.not_)
    __pos__ = _append_math_op(op.pos)
    __pow__ = _append_math_op(op.pow)
    __rshift__ = _append_math_op(op.rshift)
    __sub__ = _append_math_op(op.sub)
    __rsub__ = _append_math_op(op.sub)
    __truediv__ = _append_math_op(op.truediv)
    __xor__ = _append_math_op(op.xor)

    def __or__(lhs, rhs):
        @dispatch(argnum=0)
        def or_func(x, y):
            return x or y

        @or_func.register(jnp.ndarray)
        def lhs_array(x, y):
            if jnp.array_equal(x, jnp.array([])):
                return y if y is not None else x

            elif isinstance(y, jnp.ndarray):
                if jnp.array_equal(y, jnp.array([])):
                    return x if x is not None else y
                else:
                    return jnp.logical_or(x, y)

        return jtu.tree_map(or_func, lhs, rhs, is_leaf=lambda x: x is None)



from __future__ import annotations

import functools
import inspect
from dataclasses import dataclass

import jax

from pytreeclass.src.tree_base import explicitTreeBase, implicitTreeBase, treeBase
from pytreeclass.src.tree_indexer import treeIndexer
from pytreeclass.src.tree_op_base import treeOpBase


def treeclass(*args, **kwargs):
    """Class JAX  compaitable decorator for `dataclass`"""

    def wrapper(cls, field_only: bool):
        user_defined_init = "__init__" in cls.__dict__

        dCls = dataclass(
            unsafe_hash=True, init=not user_defined_init, repr=False, eq=False
        )(cls)

        base_classes = (dCls, treeBase)

        base_classes += (treeOpBase, treeIndexer)
        base_classes += (explicitTreeBase,) if field_only else (implicitTreeBase,)

        new_cls = type(cls.__name__, base_classes, {})

        return jax.tree_util.register_pytree_node_class(new_cls)

    if len(args) == 1 and inspect.isclass(args[0]):
        return wrapper(args[0], False)

    elif len(args) == 0 and len(kwargs) > 0:
        field_only = kwargs["field_only"] if "field_only" in kwargs else False
        return functools.partial(wrapper, field_only=field_only)



from dataclasses import field

import jax
import pytest
from jax import numpy as jnp

import pytreeclass as pytc


@pytc.treeclass
class Linear:
    weight: jnp.ndarray
    bias: jnp.ndarray

    def __init__(self, key, in_dim, out_dim):
        self.weight = jax.random.normal(key, shape=(in_dim, out_dim)) * jnp.sqrt(
            2 / in_dim
        )
        self.bias = jnp.ones((1, out_dim))

    # def __call__(self, x):
    #     return x @ self.weight + self.bias


def test_field_only():
    @pytc.treeclass(field_only=True)
    class StackedLinear:
        def __init__(self, key, in_dim, out_dim, hidden_dim):
            keys = jax.random.split(key, 3)

            # Declaring l1,l2,l3 as dataclass_fields is optional
            # as l1,l2,l3 are Linear class that is wrapped with @pytc.treeclass
            self.l1 = Linear(key=keys[0], in_dim=in_dim, out_dim=hidden_dim)
            self.l2 = Linear(key=keys[1], in_dim=hidden_dim, out_dim=hidden_dim)
            self.l3 = Linear(key=keys[2], in_dim=hidden_dim, out_dim=out_dim)

    model = StackedLinear(key=jax.random.PRNGKey(0), in_dim=2, out_dim=2, hidden_dim=2)
    model.__tree_fields__

    assert "l1" not in model.__dataclass_fields__

    @pytc.treeclass(field_only=False)
    class StackedLinear:
        def __init__(self, key, in_dim, out_dim, hidden_dim):
            keys = jax.random.split(key, 3)

            # Declaring l1,l2,l3 as dataclass_fields is optional
            # as l1,l2,l3 are Linear class that is wrapped with @pytc.treeclass
            self.l1 = Linear(key=keys[0], in_dim=in_dim, out_dim=hidden_dim)
            self.l2 = Linear(key=keys[1], in_dim=hidden_dim, out_dim=hidden_dim)
            self.l3 = Linear(key=keys[2], in_dim=hidden_dim, out_dim=out_dim)

    model = StackedLinear(key=jax.random.PRNGKey(0), in_dim=2, out_dim=2, hidden_dim=2)
    model.__tree_fields__

    assert "l1" in model.__treeclass_fields__

    @pytc.treeclass(field_only=False)
    class StackedLinear:
        l4: Linear

        def __init__(self, key, in_dim, out_dim, hidden_dim):
            keys = jax.random.split(key, 3)

            # Declaring l1,l2,l3 as dataclass_fields is optional
            # as l1,l2,l3 are Linear class that is wrapped with @pytc.treeclass
            self.l1 = Linear(key=keys[0], in_dim=in_dim, out_dim=hidden_dim)
            self.l2 = Linear(key=keys[1], in_dim=hidden_dim, out_dim=hidden_dim)
            self.l3 = Linear(key=keys[2], in_dim=hidden_dim, out_dim=out_dim)

    with pytest.raises(ValueError):
        # l4 is not declared
        model = StackedLinear(
            key=jax.random.PRNGKey(0), in_dim=2, out_dim=2, hidden_dim=2
        )
        model.__tree_fields__

    @pytc.treeclass(field_only=True)
    class StackedLinear:
        l4: Linear = field(default=pytc.static_value(1))
        l5: Linear

        def __init__(self, key, in_dim, out_dim, hidden_dim):
            ...

    with pytest.raises(ValueError):
        # l5 is not declared
        model = StackedLinear(
            key=jax.random.PRNGKey(0), in_dim=2, out_dim=2, hidden_dim=2
        )
        model.__tree_fields__


def test_hash():
    @pytc.treeclass
    class T:
        a: jnp.ndarray

    with pytest.raises(TypeError):
        hash(T(jnp.array([1, 2, 3])))



import jax
import jax.numpy as jnp

from pytreeclass import tree_viz, treeclass


def test_node():
    @treeclass
    class Linear:

        weight: jnp.ndarray
        bias: jnp.ndarray

        def __init__(self, key, in_dim, out_dim):
            self.weight = jax.random.normal(key, shape=(in_dim, out_dim)) * jnp.sqrt(
                2 / in_dim
            )
            self.bias = jnp.ones((1, out_dim))

        def __call__(self, x):
            return x @ self.weight + self.bias

    @treeclass
    class MLP:
        def __init__(self, layers, key=jax.random.PRNGKey(0)):

            keys = jax.random.split(key, len(layers))

            for i, (ki, in_dim, out_dim) in enumerate(
                zip(keys, layers[:-1], layers[1:])
            ):
                self.register_node(
                    Linear(key=ki, in_dim=in_dim, out_dim=out_dim), name=f"l{i}"
                )

        def __call__(self, x):
            x = self.l0(x)
            x = jax.nn.tanh(x)
            x = self.l1(x)
            x = jax.nn.tanh(x)
            x = self.l2(x)
            return x

    model = MLP(layers=[1, 128, 128, 1])
    x = jnp.linspace(0, 1, 100)[:, None]
    y = x**3 + jax.random.uniform(jax.random.PRNGKey(0), (100, 1)) * 0.01
    print(tree_viz.tree_diagram(model))

    # leaves,treedef=jax.tree_flatten(model)
    def loss_func(model, x, y):
        return jnp.mean((model(x) - y) ** 2)

    @jax.jit
    def update(model, x, y):
        value, grads = jax.value_and_grad(loss_func)(model, x, y)
        # no need to use `jax.tree_map` to update the model
        #  as it model is wrapped by @treeclass
        return value, model - 1e-3 * grads

    for _ in range(1, 10_001):
        value, model = update(model, x, y)

    assert value < 1e-3


from __future__ import annotations

from dataclasses import field

import jax
import jax.numpy as jnp
import jax.random as jr
import numpy as np
import numpy.testing as npt
import pytest

import pytreeclass as pytc
from pytreeclass.src.tree_indexer import _at_apply, _at_get, _at_reduce, _at_set
from pytreeclass.src.tree_util import is_treeclass_equal, static_value


@pytc.treeclass
class Test:
    a: float
    b: float
    c: float
    d: jnp.ndarray
    name: str


def test_getter_by_val():
    @pytc.treeclass
    class level1:
        a: int
        b: int
        c: int

    @pytc.treeclass
    class level2:
        d: level1
        e: level1

    A = level2(
        d=level1(a=1, b=10, c=jnp.array([1, 2, 3, 4, 5])),
        e=level1(a=2, b=20, c=jnp.array([-1, -2, -3, -4, -5])),
    )

    B = A.at[A > 0].get()
    C = level2(
        d=level1(a=1, b=10, c=jnp.array([1, 2, 3, 4, 5])),
        e=level1(a=2, b=20, c=jnp.array([])),
    )

    assert is_treeclass_equal(B, C)

    B = A.at[(A > 0) & (A < 5)].get()
    C = level2(
        d=level1(a=1, b=None, c=jnp.array([1, 2, 3, 4])),
        e=level1(a=2, b=None, c=jnp.array([])),
    )

    assert is_treeclass_equal(B, C)

    B = A.at[A == 0].get()
    C = level2(
        d=level1(a=None, b=None, c=jnp.array([])),
        e=level1(a=None, b=None, c=jnp.array([])),
    )

    assert is_treeclass_equal(B, C)

    with pytest.raises(AssertionError):
        B = A.at[A].get()

    with pytest.raises(NotImplementedError):
        B = A.at[0].get()


def test_getter_by_param():

    A = Test(10, 20, 30, jnp.array([1, 2, 3, 4, 5]), static_value("A"))

    B = A.at[A == "a"].get(array_as_leaves=False)
    assert is_treeclass_equal(B, Test(10, None, None, None, static_value("A")))

    B = A.at[(A == "a") | (A == "b")].get(array_as_leaves=False)
    assert is_treeclass_equal(B, Test(10, 20, None, None, static_value("A")))

    B = A.at[A == ""].get(array_as_leaves=False)
    assert is_treeclass_equal(B, Test(None, None, None, None, static_value("A")))

    B = A.at[(A == "a") | (A == "b") | (A == "c")].get(array_as_leaves=False)
    assert is_treeclass_equal(B, Test(10, 20, 30, None, static_value("A")))

    B = A.at[(A == "a") | (A == "b") | (A == "c") | (A == "d")].get(
        array_as_leaves=False
    )
    assert is_treeclass_equal(
        B, Test(10, 20, 30, jnp.array([1, 2, 3, 4, 5]), static_value("A"))
    )

    @pytc.treeclass
    class L0:
        a: int = 1
        b: int = 2
        c: int = 3

    @pytc.treeclass
    class L1:
        a: int = 1
        b: int = 2
        c: int = 3
        d: L0 = L0()

    @pytc.treeclass
    class L2:
        a: int = 10
        b: int = 20
        c: int = 30
        d: L1 = L1()

    t = L2()
    lhs = t.at[t == "a"].get()
    rhs = L2(10, None, None, L1(1, None, None, L0(1, None, None)))
    assert is_treeclass_equal(lhs, rhs)

    with pytest.raises(NotImplementedError):
        t.at["s"].get()

    with pytest.raises(NotImplementedError):
        t.at["s"].set(100)

    with pytest.raises(NotImplementedError):
        t.at["s"].apply(lambda _: 100)


def test_getter_by_metadata():
    @pytc.treeclass
    class Test:
        a: float = field(metadata={"name": "a", "unit": "m"})
        b: float = field(metadata={"name": "b", "unit": "m"})
        c: float = field(metadata={"name": "c", "unit": "m"})
        d: jnp.ndarray = field(metadata={"name": "d", "unit": "m"})
        name: str

    A = Test(10, 20, 30, jnp.array([1, 2, 3, 4, 5]), static_value("A"))

    B = A.at[A == {"name": "a"}].get(array_as_leaves=False)
    assert is_treeclass_equal(B, Test(10, None, None, None, static_value("A")))

    B = A.at[(A == {"name": "a"}) | (A == {"name": "b"})].get(array_as_leaves=False)
    assert is_treeclass_equal(B, Test(10, 20, None, None, static_value("A")))

    B = A.at[A == {"": ""}].get(array_as_leaves=False)
    assert is_treeclass_equal(B, Test(None, None, None, None, static_value("A")))

    B = A.at[(A == {"name": "a"}) | (A == {"name": "b"}) | (A == {"name": "c"})].get(
        array_as_leaves=False
    )
    assert is_treeclass_equal(B, Test(10, 20, 30, None, static_value("A")))

    B = A.at[
        (A == {"name": "a"})
        | (A == {"name": "b"})
        | (A == {"name": "c"})
        | (A == {"name": "d"})
    ].get(array_as_leaves=False)
    assert is_treeclass_equal(
        B, Test(10, 20, 30, jnp.array([1, 2, 3, 4, 5]), static_value("A"))
    )

    @pytc.treeclass
    class L0:
        a: int = field(default=1, metadata={"name": "a", "unit": "m"})
        b: int = 2
        c: int = 3

    @pytc.treeclass
    class L1:
        a: int = field(default=1, metadata={"name": "a", "unit": "m"})
        b: int = 2
        c: int = 3
        d: L0 = L0()

    @pytc.treeclass
    class L2:
        a: int = field(default=10, metadata={"name": "a", "unit": "m"})
        b: int = 20
        c: int = 30
        d: L1 = L1()

    t = L2()
    lhs = t.at[t == {"name": "a"}].get()
    rhs = L2(10, None, None, L1(1, None, None, L0(1, None, None)))
    assert is_treeclass_equal(lhs, rhs)


def test_setter_by_val():
    @pytc.treeclass
    class level1:
        a: int
        b: int
        c: int

    @pytc.treeclass
    class level2:
        d: level1
        e: level1

    A = level2(
        d=level1(a=1, b=10, c=jnp.array([1, 2, 3, 4, 5])),
        e=level1(a=2, b=20, c=jnp.array([-1, -2, -3, -4, -5])),
    )

    B = A.at[A < 0].set(100)
    C = level2(
        d=level1(a=1, b=10, c=jnp.array([1, 2, 3, 4, 5])),
        e=level1(a=2, b=20, c=jnp.array([100, 100, 100, 100, 100])),
    )

    assert is_treeclass_equal(B, C)

    A = Test(10, 20, 30, jnp.array([1, 2, 3, 4, 5]), static_value("A"))

    with pytest.raises(AssertionError):
        B = A.at[A].set(0)

    with pytest.raises(NotImplementedError):
        B = A.at[0].set(0)

    @pytc.treeclass
    class L0:
        a: int = 1
        b: int = 2
        c: int = 3

    @pytc.treeclass
    class L1:
        a: int = 1
        b: int = 2
        c: int = 3
        d: L0 = L0()

    @pytc.treeclass
    class L2:
        a: int = 10
        b: int = 20
        c: int = 30
        d: L1 = L1()

    t = L2()
    lhs = t.at[t == "a"].set(100)
    rhs = L2(100, 20, 30, L1(100, 2, 3, L0(100, 2, 3)))
    assert is_treeclass_equal(lhs, rhs)


def test_setter_by_param():
    A = Test(10, 20, 30, jnp.array([1, 2, 3, 4, 5]), static_value("A"))

    B = A.at[A == "a"].set(0)
    assert is_treeclass_equal(
        B, Test(0, 20, 30, jnp.array([1, 2, 3, 4, 5]), static_value("A"))
    )

    B = A.at[(A == "a") | (A == "b")].set(0)
    assert is_treeclass_equal(
        B, Test(0, 0, 30, jnp.array([1, 2, 3, 4, 5]), static_value("A"))
    )

    B = A.at[(A == "a") | (A == "b") | (A == "c")].set(0)
    assert is_treeclass_equal(
        B, Test(0, 0, 0, jnp.array([1, 2, 3, 4, 5]), static_value("A"))
    )


def test_setter_by_metadata():
    @pytc.treeclass
    class Test:
        a: float = field(metadata={"name": "a", "unit": "m"})
        b: float = field(metadata={"name": "b", "unit": "m"})
        c: float = field(metadata={"name": "c", "unit": "m"})
        d: jnp.ndarray = field(metadata={"name": "d", "unit": "m"})
        name: str

    A = Test(10, 20, 30, jnp.array([1, 2, 3, 4, 5]), static_value("A"))

    B = A.at[A == {"name": "a"}].set(100, array_as_leaves=False)
    assert is_treeclass_equal(
        B, Test(100, 20, 30, jnp.array([1, 2, 3, 4, 5]), static_value("A"))
    )

    B = A.at[(A == {"name": "a"}) | (A == {"name": "b"})].set(
        100, array_as_leaves=False
    )
    assert is_treeclass_equal(
        B, Test(100, 100, 30, jnp.array([1, 2, 3, 4, 5]), static_value("A"))
    )

    B = A.at[A == {"": ""}].set(100, array_as_leaves=False)
    assert is_treeclass_equal(
        B, Test(10, 20, 30, jnp.array([1, 2, 3, 4, 5]), static_value("A"))
    )

    B = A.at[(A == {"name": "a"}) | (A == {"name": "b"}) | (A == {"name": "c"})].set(
        100, array_as_leaves=False
    )
    assert is_treeclass_equal(
        B, Test(100, 100, 100, jnp.array([1, 2, 3, 4, 5]), static_value("A"))
    )

    B = A.at[
        (A == {"name": "a"})
        | (A == {"name": "b"})
        | (A == {"name": "c"})
        | (A == {"name": "d"})
    ].set(100, array_as_leaves=False)
    assert is_treeclass_equal(B, Test(100, 100, 100, 100, static_value("A")))

    @pytc.treeclass
    class L0:
        a: int = field(default=1, metadata={"name": "a", "unit": "m"})
        b: int = 2
        c: int = 3

    @pytc.treeclass
    class L1:
        a: int = field(default=1, metadata={"name": "a", "unit": "m"})
        b: int = 2
        c: int = 3
        d: L0 = L0()

    @pytc.treeclass
    class L2:
        a: int = field(default=10, metadata={"name": "a", "unit": "m"})
        b: int = 20
        c: int = 30
        d: L1 = L1()

    t = L2()
    lhs = t.at[t == {"name": "a"}].set(100)
    rhs = L2(100, 20, 30, L1(100, 2, 3, L0(100, 2, 3)))
    assert is_treeclass_equal(lhs, rhs)

    @pytc.treeclass
    class T:
        a: jnp.ndarray

    t = T(True)
    assert is_treeclass_equal(t.at[t == bool].set(False), T(a=False))


def test_apply_and_its_derivatives():
    @pytc.treeclass
    class A:
        a: int
        b: int
        c: jnp.ndarray

    init = A(1, 2, jnp.array([1, 2, 3, 4, 5]))

    # By boolean pytree
    lhs = A(1, 4, jnp.array([1, 4, 9, 16, 25]))
    rhs = init.at[init == init].apply(lambda x: x**2)
    assert is_treeclass_equal(lhs, rhs)

    lhs = A(1, 4, jnp.array([1, 4, 9, 16, 25]))
    rhs = init.at[...].apply(lambda x: x**2)
    assert is_treeclass_equal(lhs, rhs)

    lhs = A(2, 3, jnp.array([2, 3, 4, 5, 6]))
    rhs = init.at[init == init].apply(lambda x: x + 1)
    assert is_treeclass_equal(lhs, rhs)

    lhs = A(2, 3, jnp.array([2, 3, 4, 5, 6]))
    rhs = init.at[...].apply(lambda x: x + 1)
    assert is_treeclass_equal(lhs, rhs)

    lhs = A(20, 30, jnp.array([20, 30, 40, 50, 60]))
    rhs = init.at[init == init].apply(lambda x: (x + 1) * 10)
    assert is_treeclass_equal(lhs, rhs)

    lhs = A(20, 30, jnp.array([20, 30, 40, 50, 60]))
    rhs = init.at[...].apply(lambda x: (x + 1) * 10)
    assert is_treeclass_equal(lhs, rhs)

    @pytc.treeclass
    class L0:
        a: int = 1
        b: int = 2
        c: int = 3

    @pytc.treeclass
    class L1:
        a: int = 1
        b: int = 2
        c: int = 3
        d: L0 = L0()

    @pytc.treeclass
    class L2:
        a: int = 10
        b: int = 20
        c: int = 30
        d: L1 = L1()

    t = L2()
    lhs = t.at[t == "a"].apply(lambda _: 100)
    rhs = L2(100, 20, 30, L1(100, 2, 3, L0(100, 2, 3)))
    assert is_treeclass_equal(lhs, rhs)

    lhs = A(2, 3, jnp.array([2, 3, 4, 5, 6]))
    rhs = init.at[init == init].add(1)
    assert is_treeclass_equal(lhs, rhs)

    lhs = A(2, 3, jnp.array([2, 3, 4, 5, 6]))
    rhs = init.at[...].add(1)
    assert is_treeclass_equal(lhs, rhs)

    lhs = A(0.5, 1.0, jnp.array([0.5, 1.0, 1.5, 2.0, 2.5]))
    rhs = init.at[init == init].divide(2.0)
    assert is_treeclass_equal(lhs, rhs)

    lhs = A(0.5, 1.0, jnp.array([0.5, 1.0, 1.5, 2.0, 2.5]))
    rhs = init.at[...].divide(2.0)
    assert is_treeclass_equal(lhs, rhs)

    lhs = A(1, 1, jnp.array([1, 1, 1, 1, 1]))
    rhs = init.at[init == init].min(1)
    assert is_treeclass_equal(lhs, rhs)

    lhs = A(4, 4, jnp.array([4, 4, 4, 4, 5]))
    rhs = init.at[init == init].max(4)
    assert is_treeclass_equal(lhs, rhs)

    lhs = A(4, 4, jnp.array([4, 4, 4, 4, 5]))
    rhs = init.at[...].max(4)
    assert is_treeclass_equal(lhs, rhs)

    lhs = A(1, 4, jnp.array([1, 4, 9, 16, 25]))
    rhs = init.at[init == init].power(2)
    assert is_treeclass_equal(lhs, rhs)

    lhs = A(1, 4, jnp.array([1, 4, 9, 16, 25]))
    rhs = init.at[...].power(2)
    assert is_treeclass_equal(lhs, rhs)

    lhs = A(1, 2, jnp.array([1, 2, 3, 16, 25]))
    rhs = init.at[init > 3].apply(lambda x: x**2)
    assert is_treeclass_equal(lhs, rhs)

    lhs = A(2, 3, jnp.array([2, 3, 4, 5, 6]))
    rhs = init.at[init > 0].apply(lambda x: x + 1)
    assert is_treeclass_equal(lhs, rhs)

    rhs = init.at[init > 100].apply(lambda x: (x + 1) * 10)
    assert is_treeclass_equal(init, rhs)

    lhs = A(2, 3, jnp.array([2, 3, 4, 5, 6]))
    rhs = init.at[init > 0].add(1)
    assert is_treeclass_equal(lhs, rhs)

    lhs = A(1, 4, jnp.array([1, 4, 3, 4, 5]))
    rhs = init.at[init == 2].multiply(2)
    assert is_treeclass_equal(lhs, rhs)

    # by param

    lhs = A(1, 2, jnp.array([1, 2, 3, 4, 5]))
    rhs = init.at[init == "a"].apply(lambda x: x**2)
    assert is_treeclass_equal(lhs, rhs)

    lhs = A(2, 2, jnp.array([1, 2, 3, 4, 5]))
    rhs = init.at[init == "a"].apply(lambda x: x + 1)
    assert is_treeclass_equal(lhs, rhs)

    lhs = A(20, 2, jnp.array([1, 2, 3, 4, 5]))
    rhs = init.at[init == "a"].apply(lambda x: (x + 1) * 10)
    assert is_treeclass_equal(lhs, rhs)

    lhs = A(2, 2, jnp.array([1, 2, 3, 4, 5]))
    rhs = init.at[init == "a"].add(1)
    assert is_treeclass_equal(lhs, rhs)

    lhs = A(0.5, 2, jnp.array([1, 2, 3, 4, 5]))
    rhs = init.at[init == "a"].divide(2.0)
    assert is_treeclass_equal(lhs, rhs)

    lhs = A(1, 2, jnp.array([1, 2, 3, 4, 5]))
    rhs = init.at[init == "a"].min(1)
    assert is_treeclass_equal(lhs, rhs)

    lhs = A(4, 2, jnp.array([1, 2, 3, 4, 5]))
    rhs = init.at[init == "a"].max(4)
    assert is_treeclass_equal(lhs, rhs)

    lhs = A(1, 2, jnp.array([1, 2, 3, 4, 5]))
    rhs = init.at[init == "a"].power(2)
    assert is_treeclass_equal(lhs, rhs)

    lhs = A(2, 2, jnp.array([1, 2, 3, 4, 5]))
    rhs = init.at[init == "a"].multiply(2)
    assert is_treeclass_equal(lhs, rhs)

    #
    lhs = A(1, 4, jnp.array([1, 4, 9, 16, 25]))
    rhs = init.at[init != "a"].apply(lambda x: x**2)
    assert is_treeclass_equal(lhs, rhs)

    lhs = A(1, 3, jnp.array([2, 3, 4, 5, 6]))
    rhs = init.at[init != "a"].apply(lambda x: x + 1)
    assert is_treeclass_equal(lhs, rhs)

    lhs = A(1, 30, jnp.array([20, 30, 40, 50, 60]))
    rhs = init.at[init != "a"].apply(lambda x: (x + 1) * 10)
    assert is_treeclass_equal(lhs, rhs)

    lhs = A(1, 3, jnp.array([2, 3, 4, 5, 6]))
    rhs = init.at[init != "a"].add(1)
    assert is_treeclass_equal(lhs, rhs)

    lhs = A(1, 1, jnp.array([0.5, 1, 1.5, 2, 2.5]))
    rhs = init.at[init != "a"].divide(2.0)
    assert is_treeclass_equal(lhs, rhs)

    # raise
    with pytest.raises(ValueError):
        init.freeze().at[init == "a"].apply(lambda x: x**2)

    with pytest.raises(ValueError):
        init.freeze().at[(init == "a") | (A == "b")].apply(lambda x: x**2)

    with pytest.raises(ValueError):
        init.freeze().at[(init == "a") | (A == "b")].set(0)

    with pytest.raises(ValueError):
        init.freeze().at[(init == init) | (A == "b")].set(0)

    with pytest.raises(ValueError):
        init.freeze().at[(init == init) | (A == "b")].reduce_sum()

    @pytc.treeclass
    class Test:
        a: float = field(metadata={"name": "a", "unit": "m"})
        b: float = field(metadata={"name": "b", "unit": "m"})
        c: float = field(metadata={"name": "c", "unit": "m"})
        d: jnp.ndarray = field(metadata={"name": "d", "unit": "m"})
        name: str

    A = Test(10, 20, 30, jnp.array([1, 2, 3, 4, 5]), static_value("A"))

    B = A.at[A == {"name": "a"}].apply(lambda _: 100, array_as_leaves=False)
    assert is_treeclass_equal(
        B, Test(100, 20, 30, jnp.array([1, 2, 3, 4, 5]), static_value("A"))
    )

    B = A.at[(A == {"name": "a"}) | (A == {"name": "b"})].set(
        100, array_as_leaves=False
    )
    assert is_treeclass_equal(
        B, Test(100, 100, 30, jnp.array([1, 2, 3, 4, 5]), static_value("A"))
    )

    B = A.at[A == {"": ""}].apply(lambda _: 100, array_as_leaves=False)
    assert is_treeclass_equal(
        B, Test(10, 20, 30, jnp.array([1, 2, 3, 4, 5]), static_value("A"))
    )

    B = A.at[(A == {"name": "a"}) | (A == {"name": "b"}) | (A == {"name": "c"})].set(
        100, array_as_leaves=False
    )
    assert is_treeclass_equal(
        B, Test(100, 100, 100, jnp.array([1, 2, 3, 4, 5]), static_value("A"))
    )

    B = A.at[
        (A == {"name": "a"})
        | (A == {"name": "b"})
        | (A == {"name": "c"})
        | (A == {"name": "d"})
    ].apply(lambda _: 100, array_as_leaves=False)
    assert is_treeclass_equal(B, Test(100, 100, 100, 100, static_value("A")))

    B = A.at[A != {"name": "a"}].apply(lambda _: 100)
    assert is_treeclass_equal(
        B, Test(10, 100, 100, jnp.array([100, 100, 100, 100, 100]), static_value("A"))
    )

    @pytc.treeclass
    class L0:
        a: int = field(default=1, metadata={"name": "a", "unit": "m"})
        b: int = 2
        c: int = 3

    @pytc.treeclass
    class L1:
        a: int = field(default=1, metadata={"name": "a", "unit": "m"})
        b: int = 2
        c: int = 3
        d: L0 = L0()

    @pytc.treeclass
    class L2:
        a: int = field(default=10, metadata={"name": "a", "unit": "m"})
        b: int = 20
        c: int = 30
        d: L1 = L1()

    t = L2()
    lhs = t.at[t == {"name": "a"}].apply(lambda _: 100)
    rhs = L2(100, 20, 30, L1(100, 2, 3, L0(100, 2, 3)))
    assert is_treeclass_equal(lhs, rhs)

    t = L2()
    lhs = t.at[t != {"name": "a"}].apply(lambda _: 100)
    rhs = L2(10, 100, 100, L1(1, 100, 100, L0(1, 100, 100)))

    assert is_treeclass_equal(lhs, rhs)

    @pytc.treeclass
    class L0:
        a: int = 1
        b: int = 2
        c: int = 3

    @pytc.treeclass
    class L1:
        a: int = 1
        b: int = 2
        c: int = 3
        d: L0 = L0()

    @pytc.treeclass
    class L2:
        a: int = 10
        b: int = 20
        c: int = 30
        d: L1 = field(default=L1(), metadata={"name": "d", "unit": "m"})

    t = L2()
    lhs = t.at[t == {"name": "d"}].apply(lambda _: 100)
    rhs = L2(10, 20, 30, L1(100, 100, 100, L0(100, 100, 100)))
    assert is_treeclass_equal(lhs, rhs)


def test_reduce():
    @pytc.treeclass
    class A:
        a: int
        b: int
        c: jnp.ndarray

    init = A(1, 2, jnp.array([1, 2, 3, 4, 5]))

    lhs = 2 + 2 + 3 + 4 + 5
    rhs = init.at[init > 1].reduce(lambda x, y: x + jnp.sum(y))
    assert lhs == rhs

    lhs = 3 + 4 + 5
    rhs = init.at[init > 2].reduce(lambda x, y: x + jnp.sum(y))
    assert lhs == rhs

    lhs = 0
    rhs = init.at[init > 100].reduce(lambda x, y: x + jnp.sum(y))
    assert lhs == rhs

    @pytc.treeclass
    class B:
        a: int
        b: int
        c: jnp.ndarray
        d: tuple

    init = B(1, 2, jnp.array([1, 2, 3, 4, 5]), (10, 20, 30))

    lhs = 2 + 2 + 3 + 4 + 5 + 10 + 20 + 30
    rhs = init.at[init > 1].reduce(lambda x, y: x + jnp.sum(y))
    assert lhs == rhs

    with pytest.raises(AssertionError):
        init.at[init].reduce(lambda x, y: x + jnp.sum(y))


def test_reduce_and_its_derivatives():
    @pytc.treeclass
    class Linear:
        weight: jnp.ndarray
        bias: jnp.ndarray

        def __init__(self, key, in_dim, out_dim):
            self.weight = jax.random.normal(key, shape=(in_dim, out_dim)) * jnp.sqrt(
                2 / in_dim
            )
            self.bias = jnp.ones((1, out_dim))

        # def __call__(self, x):
        #     return x @ self.weight + self.bias

    @pytc.treeclass
    class StackedLinear:
        l1: Linear = field(metadata={"description": "First layer"})
        l2: Linear = field(metadata={"description": "Second layer"})

        def __init__(self, key, in_dim, out_dim, hidden_dim):
            keys = jax.random.split(key, 3)

            self.l1 = Linear(key=keys[0], in_dim=in_dim, out_dim=hidden_dim)
            self.l2 = Linear(key=keys[2], in_dim=hidden_dim, out_dim=out_dim)

        # def __call__(self, x):
        #     x = self.l1(x)
        #     x = jax.nn.tanh(x)
        #     x = self.l2(x)

        #     return x

    model = StackedLinear(in_dim=1, out_dim=1, hidden_dim=5, key=jax.random.PRNGKey(0))

    assert (model.at[model > 0].reduce_min()) == 0.98507565
    assert (model.at[model > 0].reduce_max()) == 1.3969219
    assert (model.at[model > 0].reduce_sum()) == 10.6970625
    assert (model.at[model > 0].reduce_product()) == 1.8088213

    assert (model.at[model == "l1"].reduce_sum()) == 2.8428133
    assert (model.at[model == "l1"].reduce_product()) == -3.4602268

    assert (model.at[model == Linear].reduce_min()) == -2.8383057
    assert (model.at[model == Linear].reduce_max()) == 1.3969219
    assert (model.at[model == Linear].reduce_sum()) == 3.3538322
    assert (model.at[model == Linear].reduce_product()) == 0.84782064

    assert (model.at[model == jnp.ndarray].reduce_min()) == -2.8383057
    assert (model.at[model == jnp.ndarray].reduce_max()) == 1.3969219
    assert (model.at[model == jnp.ndarray].reduce_sum()) == 3.3538322
    assert (model.at[model == jnp.ndarray].reduce_product()) == 0.84782064


def test_not_implemented():
    @pytc.treeclass
    class Test:
        a: int = 1

    t = Test()

    with pytest.raises(NotImplementedError):
        _at_get(t == t, 3)

    with pytest.raises(NotImplementedError):
        _at_set(t == t, 3, 3)

    with pytest.raises(NotImplementedError):
        _at_reduce(t == t, lambda x: x, 3)

    with pytest.raises(NotImplementedError):
        _at_apply(t == t, lambda x: x, 3)

    with pytest.raises(NotImplementedError):
        _at_get(t == t, np.array([1, 2, 3]))

    with pytest.raises(NotImplementedError):
        _at_set(t == t, 3, np.array([1, 2, 3]))

    with pytest.raises(NotImplementedError):
        _at_reduce(t == t, lambda x: x, np.array([1, 2, 3]))

    with pytest.raises(NotImplementedError):
        _at_apply(t == t, lambda x: x, np.array([1, 2, 3]))


def test_is_leaf():
    @pytc.treeclass
    class Test:
        a: int

    t = Test([1, 2, 3, None])

    assert is_treeclass_equal(
        t.at[...].set(10, is_leaf=lambda x: x is None), Test([10, 10, 10, 10])
    )

    assert is_treeclass_equal(
        t.at[...].apply(lambda x: 10, is_leaf=lambda x: x is None),
        Test([10, 10, 10, 10]),
    )


def test_masking():
    @pytc.treeclass
    class Linear:
        weight: jnp.ndarray
        bias: jnp.ndarray

        def __init__(self, key, in_dim, out_dim):
            self.weight = jax.random.normal(key, shape=(in_dim, out_dim)) * jnp.sqrt(
                2 / in_dim
            )
            self.bias = jnp.ones((1, out_dim))

        def __call__(self, x):
            return x @ self.weight + self.bias

    @pytc.treeclass
    class Dropout:
        p: float
        eval: bool | None

        def __init__(self, p: float = 0.5, eval: bool | None = None):
            """p : probability of an element to be zeroed out"""
            self.p = p
            self.eval = eval

        def __call__(self, x, *, key=jr.PRNGKey(0)):
            return (
                x
                if (self.eval is True)
                else jnp.where(
                    jr.bernoulli(key, (1 - self.p), x.shape), x / (1 - self.p), 0
                )
            )

    @pytc.treeclass
    class LinearWithDropout:
        def __init__(self):
            self.l1 = Linear(key=jr.PRNGKey(0), in_dim=1, out_dim=5)
            self.d1 = Dropout(p=1.0)  # zero out all elements

        def __call__(self, x):
            x = self.l1(x)
            x = self.d1(x)
            return x

    model = LinearWithDropout()
    npt.assert_allclose(model(jnp.ones((1, 1))), jnp.array([[0.0, 0.0, 0.0, 0.0, 0.0]]))

    mask = model == "eval"
    model_no_dropout = model.at[mask].set(True, is_leaf=lambda x: x is None)
    npt.assert_allclose(
        model_no_dropout(jnp.ones((1, 1))),
        jnp.array([[1.2656513, -0.8149204, 0.61661845, 2.7664368, 1.3457328]]),
    )



from typing import Callable, Sequence

import jax
import jax.numpy as jnp
import numpy as np

from pytreeclass import treeclass


def test_nn():
    @treeclass
    class Linear:

        weight: jnp.ndarray
        bias: jnp.ndarray

        def __init__(self, key, in_dim, out_dim):
            self.weight = jax.random.normal(
                key, shape=(in_dim, out_dim)) * jnp.sqrt(2 / in_dim)  # fmt: skip
            self.bias = jnp.ones((1, out_dim))

        def __call__(self, x):
            return x @ self.weight + self.bias

    @treeclass
    class StackedLinear:
        layers: Sequence[Linear]

        def __init__(self, key, layers):
            keys = jax.random.split(key, len(layers) - 1)

            self.layers = []

            for ki, in_dim, out_dim in zip(keys, layers[:-1], layers[1:]):
                self.layers += [Linear(ki, in_dim, out_dim)]

        def __call__(self, x):
            for layer in self.layers[:-1]:
                x = layer(x)
                x = jax.nn.tanh(x)

            return self.layers[-1](x)

    x = jnp.linspace(0, 1, 100)[:, None]
    y = x**3 + jax.random.uniform(jax.random.PRNGKey(0), (100, 1)) * 0.01

    model = StackedLinear(layers=[1, 128, 128, 1], key=jax.random.PRNGKey(0))

    def loss_func(model, x, y):
        return jnp.mean((model(x) - y) ** 2)

    @jax.jit
    def update(model, x, y):
        value, grads = jax.value_and_grad(loss_func)(model, x, y)
        return value, jax.tree_map(lambda x, y: x - 1e-3 * y, model, grads)

    for _ in range(1, 2001):
        value, model = update(model, x, y)

    np.testing.assert_allclose(value, jnp.array(0.00103019), atol=1e-5)


def test_nn_with_func_input():
    @treeclass
    class Linear:

        weight: jnp.ndarray
        bias: jnp.ndarray
        act_func: Callable

        def __init__(self, key, in_dim, out_dim, act_func):
            self.act_func = act_func
            self.weight = jax.random.normal(key, shape=(in_dim, out_dim)) * jnp.sqrt(
                2 / in_dim
            )
            self.bias = jnp.ones((1, out_dim))

        def __call__(self, x):
            return self.act_func(x @ self.weight + self.bias)

    layer = Linear(key=jax.random.PRNGKey(0), in_dim=1, out_dim=1, act_func=jax.nn.tanh)
    x = jnp.linspace(0, 1, 100)[:, None]
    # y = x**3 + jax.random.uniform(jax.random.PRNGKey(0), (100, 1)) * 0.01
    layer(x)
    return True


def test_compact_nn():
    @treeclass
    class Linear:

        weight: jnp.ndarray
        bias: jnp.ndarray

        def __init__(self, key, in_dim, out_dim):
            self.weight = jax.random.normal(key, shape=(in_dim, out_dim)) * jnp.sqrt(
                2 / in_dim
            )
            self.bias = jnp.ones((1, out_dim))

        def __call__(self, x):
            return x @ self.weight + self.bias

    @treeclass
    class StackedLinear:
        def __init__(self, key, in_dim, out_dim, hidden_dim):

            keys = jax.random.split(key, 3)

            self.l1 = Linear(key=keys[0], in_dim=in_dim, out_dim=hidden_dim)
            self.l2 = Linear(key=keys[1], in_dim=hidden_dim, out_dim=hidden_dim)
            self.l3 = Linear(key=keys[2], in_dim=hidden_dim, out_dim=out_dim)

        def __call__(self, x):
            x = self.l1(x)
            x = jax.nn.tanh(x)
            x = self.l2(x)
            x = jax.nn.tanh(x)
            x = self.l3(x)

            return x

    x = jnp.linspace(0, 1, 100)[:, None]
    y = x**3 + jax.random.uniform(jax.random.PRNGKey(0), (100, 1)) * 0.01

    model = StackedLinear(in_dim=1, out_dim=1, hidden_dim=10, key=jax.random.PRNGKey(0))

    def loss_func(model, x, y):
        return jnp.mean((model(x) - y) ** 2)

    @jax.jit
    def update(model, x, y):
        value, grads = jax.value_and_grad(loss_func)(model, x, y)

        # no need to use `jax.tree_map` to update the model
        #  as it model is wrapped by @treeclass
        return value, model - 1e-3 * grads

    for _ in range(1, 10_001):
        value, model = update(model, x, y)

    np.testing.assert_allclose(value, jnp.array(0.0031012), atol=1e-5)



from .src import tree_util, tree_viz
from .src.decorator import treeclass
from .src.tree_util import static_field, static_value

__all__ = (
    "tree_viz",
    "treeclass",
    "static_value",
    "static_field",
    "tree_viz",
    "tree_util",
)

__version__ = "0.0.9post1"



from dataclasses import field

import jax
import jax.numpy as jnp
import numpy as np
import pytest

import pytreeclass as pytc
from pytreeclass.src.tree_util import (
    _freeze_nodes,
    _unfreeze_nodes,
    is_treeclass_equal,
    static_value,
)


def test_freezing_unfreezing():
    @pytc.treeclass
    class A:
        a: int
        b: int

    a = A(1, 2)
    b = a.freeze()
    c = a.unfreeze()

    assert jax.tree_util.tree_leaves(a) == [1, 2]
    assert jax.tree_util.tree_leaves(b) == []
    assert jax.tree_util.tree_leaves(c) == [1, 2]

    a = A(1, 2)
    b = a.at[...].static()

    assert jax.tree_util.tree_leaves(a) == [1, 2]
    assert jax.tree_util.tree_leaves(b) == []

    @pytc.treeclass
    class A:
        a: int
        b: int

    @pytc.treeclass
    class B:
        c: int = 3
        d: A = A(1, 2)

    a = B()
    a.d = a.d.freeze()
    assert a.d.frozen is True
    assert (
        pytc.tree_viz.tree_diagram(a)
    ) == "B\n    ├── c=3\n    └#─ d=A\n        ├#─ a=1\n        └#─ b=2     "

    @pytc.treeclass(field_only=True)
    class A:
        a: int
        b: int

    a = A(1, 2)
    b = a.freeze()
    c = a.unfreeze()

    assert jax.tree_util.tree_leaves(a) == [1, 2]
    assert jax.tree_util.tree_leaves(b) == []
    assert jax.tree_util.tree_leaves(c) == [1, 2]

    @pytc.treeclass
    class A:
        a: int
        b: int

    @pytc.treeclass
    class B:
        c: int = 3
        d: A = A(1, 2)

    a = B()
    a.d = a.d.freeze()
    assert a.d.frozen is True
    assert (
        pytc.tree_viz.tree_diagram(a)
    ) == "B\n    ├── c=3\n    └#─ d=A\n        ├#─ a=1\n        └#─ b=2     "
    assert (
        a.tree_diagram()
    ) == "B\n    ├── c=3\n    └#─ d=A\n        ├#─ a=1\n        └#─ b=2     "

    @pytc.treeclass
    class Linear:

        weight: jnp.ndarray
        bias: jnp.ndarray
        notes: str = field(default=pytc.static_value("string"))

        def __init__(self, key, in_dim, out_dim):
            self.weight = jax.random.normal(key, shape=(in_dim, out_dim)) * jnp.sqrt(
                2 / in_dim
            )
            self.bias = jnp.ones((1, out_dim))

        def __call__(self, x):
            return x @ self.weight + self.bias

    @pytc.treeclass
    class StackedLinear:
        def __init__(self, key, in_dim, out_dim, hidden_dim):

            keys = jax.random.split(key, 3)

            self.l1 = Linear(key=keys[0], in_dim=in_dim, out_dim=hidden_dim)
            self.l2 = Linear(key=keys[1], in_dim=hidden_dim, out_dim=hidden_dim)
            self.l3 = Linear(key=keys[2], in_dim=hidden_dim, out_dim=out_dim)

        def __call__(self, x):
            x = self.l1(x)
            x = jax.nn.tanh(x)
            x = self.l2(x)
            x = jax.nn.tanh(x)
            x = self.l3(x)
            return x

    x = jnp.linspace(0, 1, 100)[:, None]
    y = x**3 + jax.random.uniform(jax.random.PRNGKey(0), (100, 1)) * 0.01

    model = StackedLinear(in_dim=1, out_dim=1, hidden_dim=10, key=jax.random.PRNGKey(0))

    model = model.freeze()

    def loss_func(model, x, y):
        return jnp.mean((model(x) - y) ** 2)

    @jax.jit
    def update(model, x, y):
        value, grads = jax.value_and_grad(loss_func)(model, x, y)
        return value, model - 1e-3 * grads

    for _ in range(1, 10_001):
        value, model = update(model, x, y)

    np.testing.assert_allclose(value, jnp.array(3.9368), atol=1e-3)

    @pytc.treeclass
    class Stacked:
        def __init__(self):
            self.mdl1 = StackedLinear(
                in_dim=1, out_dim=1, hidden_dim=10, key=jax.random.PRNGKey(0)
            )
            self.mdl2 = StackedLinear(
                in_dim=1, out_dim=1, hidden_dim=10, key=jax.random.PRNGKey(0)
            )

    mdl = Stacked()

    frozen_diagram = pytc.tree_viz.tree_diagram((_freeze_nodes(mdl)))
    # trunk-ignore(flake8/E501)
    fmt = "Stacked\n    ├#─ mdl1=StackedLinear\n    │   ├#─ l1=Linear\n    │   │   ├#─ weight=f32[1,10]\n    │   │   ├#─ bias=f32[1,10]\n    │   │   └#─ notes=*'string'    \n    │   ├#─ l2=Linear\n    │   │   ├#─ weight=f32[10,10]\n    │   │   ├#─ bias=f32[1,10]\n    │   │   └#─ notes=*'string'    \n    │   └#─ l3=Linear\n    │       ├#─ weight=f32[10,1]\n    │       ├#─ bias=f32[1,1]\n    │       └#─ notes=*'string'        \n    └#─ mdl2=StackedLinear\n        ├#─ l1=Linear\n        │   ├#─ weight=f32[1,10]\n        │   ├#─ bias=f32[1,10]\n        │   └#─ notes=*'string'    \n        ├#─ l2=Linear\n        │   ├#─ weight=f32[10,10]\n        │   ├#─ bias=f32[1,10]\n        │   └#─ notes=*'string'    \n        └#─ l3=Linear\n            ├#─ weight=f32[10,1]\n            ├#─ bias=f32[1,1]\n            └#─ notes=*'string'            "

    unfrozen_diagram = pytc.tree_viz.tree_diagram(_unfreeze_nodes(_freeze_nodes(mdl)))
    # trunk-ignore(flake8/E501)
    fmt = "Stacked\n    ├── mdl1=StackedLinear\n    │   ├── l1=Linear\n    │   │   ├── weight=f32[1,10]\n    │   │   ├── bias=f32[1,10]\n    │   │   └── notes=*'string' \n    │   ├── l2=Linear\n    │   │   ├── weight=f32[10,10]\n    │   │   ├── bias=f32[1,10]\n    │   │   └── notes=*'string' \n    │   └── l3=Linear\n    │       ├── weight=f32[10,1]\n    │       ├── bias=f32[1,1]\n    │       └── notes=*'string'     \n    └── mdl2=StackedLinear\n        ├── l1=Linear\n        │   ├── weight=f32[1,10]\n        │   ├── bias=f32[1,10]\n        │   └── notes=*'string' \n        ├── l2=Linear\n        │   ├── weight=f32[10,10]\n        │   ├── bias=f32[1,10]\n        │   └── notes=*'string' \n        └── l3=Linear\n            ├── weight=f32[10,1]\n            ├── bias=f32[1,1]\n            └── notes=*'string'         "

    assert fmt == unfrozen_diagram

    @pytc.treeclass(field_only=True)
    class StackedLinear:
        l1: Linear
        l2: Linear
        l3: Linear

        def __init__(self, key, in_dim, out_dim, hidden_dim):

            keys = jax.random.split(key, 3)

            self.l1 = Linear(key=keys[0], in_dim=in_dim, out_dim=hidden_dim)
            self.l2 = Linear(key=keys[1], in_dim=hidden_dim, out_dim=hidden_dim)
            self.l3 = Linear(key=keys[2], in_dim=hidden_dim, out_dim=out_dim)

        def __call__(self, x):
            x = self.l1(x)
            x = jax.nn.tanh(x)
            x = self.l2(x)
            x = jax.nn.tanh(x)
            x = self.l3(x)
            return x

    x = jnp.linspace(0, 1, 100)[:, None]
    y = x**3 + jax.random.uniform(jax.random.PRNGKey(0), (100, 1)) * 0.01

    model = StackedLinear(in_dim=1, out_dim=1, hidden_dim=10, key=jax.random.PRNGKey(0))

    model = model.freeze()

    @jax.jit
    def update(model, x, y):
        value, grads = jax.value_and_grad(loss_func)(model, x, y)
        return value, model - 1e-3 * grads

    for _ in range(1, 10_001):
        value, model = update(model, x, y)

    np.testing.assert_allclose(value, jnp.array(3.9368), atol=1e-3)

    @pytc.treeclass(field_only=True)
    class Stacked:
        mdl1: StackedLinear
        mdl2: StackedLinear

        def __init__(self):
            self.mdl1 = StackedLinear(
                in_dim=1, out_dim=1, hidden_dim=10, key=jax.random.PRNGKey(0)
            )
            self.mdl2 = StackedLinear(
                in_dim=1, out_dim=1, hidden_dim=10, key=jax.random.PRNGKey(0)
            )

    mdl = Stacked()

    frozen_diagram = pytc.tree_viz.tree_diagram((_freeze_nodes(mdl)))
    # trunk-ignore(flake8/E501)
    fmt = "Stacked\n    ├#─ mdl1=StackedLinear\n    │   ├#─ l1=Linear\n    │   │   ├#─ weight=f32[1,10]\n    │   │   ├#─ bias=f32[1,10]\n    │   │   └#─ notes=*'string' \n    │   ├#─ l2=Linear\n    │   │   ├#─ weight=f32[10,10]\n    │   │   ├#─ bias=f32[1,10]\n    │   │   └#─ notes=*'string' \n    │   └#─ l3=Linear\n    │       ├#─ weight=f32[10,1]\n    │       ├#─ bias=f32[1,1]\n    │       └#─ notes=*'string'     \n    └#─ mdl2=StackedLinear\n        ├#─ l1=Linear\n        │   ├#─ weight=f32[1,10]\n        │   ├#─ bias=f32[1,10]\n        │   └#─ notes=*'string' \n        ├#─ l2=Linear\n        │   ├#─ weight=f32[10,10]\n        │   ├#─ bias=f32[1,10]\n        │   └#─ notes=*'string' \n        └#─ l3=Linear\n            ├#─ weight=f32[10,1]\n            ├#─ bias=f32[1,1]\n            └#─ notes=*'string'         "
    assert fmt == frozen_diagram

    unfrozen_diagram = pytc.tree_viz.tree_diagram(_unfreeze_nodes(_freeze_nodes(mdl)))
    # trunk-ignore(flake8/E501)
    fmt = "Stacked\n    ├── mdl1=StackedLinear\n    │   ├── l1=Linear\n    │   │   ├── weight=f32[1,10]\n    │   │   ├── bias=f32[1,10]\n    │   │   └── notes=*'string' \n    │   ├── l2=Linear\n    │   │   ├── weight=f32[10,10]\n    │   │   ├── bias=f32[1,10]\n    │   │   └── notes=*'string' \n    │   └── l3=Linear\n    │       ├── weight=f32[10,1]\n    │       ├── bias=f32[1,1]\n    │       └── notes=*'string'     \n    └── mdl2=StackedLinear\n        ├── l1=Linear\n        │   ├── weight=f32[1,10]\n        │   ├── bias=f32[1,10]\n        │   └── notes=*'string' \n        ├── l2=Linear\n        │   ├── weight=f32[10,10]\n        │   ├── bias=f32[1,10]\n        │   └── notes=*'string' \n        └── l3=Linear\n            ├── weight=f32[10,1]\n            ├── bias=f32[1,1]\n            └── notes=*'string'         "
    assert fmt == unfrozen_diagram

    @pytc.treeclass(field_only=False)
    class Test:
        a: int = 1
        b: float = field(default=pytc.static_value(1.0))
        c: str = "test"

    t = Test()

    with pytest.raises(ValueError):
        t.freeze().a = 1

    @pytc.treeclass(field_only=True)
    class Test:
        a: int = 1
        b: float = field(default=static_value(1.0))
        c: str = field(default=static_value("test"))

    t = Test()
    assert jax.tree_util.tree_leaves(t) == [1]

    with pytest.raises(ValueError):
        t.freeze().a = 1

    t.unfreeze().a = 1

    hash(t)

    t = Test()
    t.unfreeze()
    t.freeze()
    assert t.frozen is False

    @pytc.treeclass
    class Test:
        a: int

    t = Test(100).freeze()

    with pytest.raises(ValueError):
        t.at[...].set(0)

    with pytest.raises(ValueError):
        t.at[...].apply(lambda x: x + 1)

    with pytest.raises(ValueError):
        t.at[...].reduce(jnp.sin)

    with pytest.raises(ValueError):
        t.at[...].static()

    class T:
        pass

    t = Test(T())

    with pytest.raises(NotImplementedError):
        t.at[...].set(0)

    with pytest.raises(NotImplementedError):
        t.at[...].apply(jnp.sin)

    with pytest.raises(NotImplementedError):
        t.at[...].reduce(jnp.sin)

    with pytest.raises(NotImplementedError):
        t.at[...].static()

    @pytc.treeclass
    class Test:
        x: jnp.ndarray

        def __init__(self, x):
            self.x = x

    t = Test(jnp.array([1, 2, 3]))
    assert is_treeclass_equal(t.at[...].set(None), Test(x=None))



import pytest

from pytreeclass.src.decorator_util import cached_property, dispatch


def test_cached_property():
    class test:
        def __init__(self, counter):
            self.counter = counter

        @cached_property
        def counter_cached_property(self):
            return self.counter

        @property
        def counter_property(self):
            return self.counter

    a = test(counter=10)
    assert a.counter_property == 10
    assert a.counter_cached_property == 10
    a.counter += 1
    assert a.counter_property == 11
    assert a.counter_cached_property == 10


def test_dispatch():
    @dispatch(argnum=0)
    def func(x):
        raise ValueError("unknown")

    @func.register(int)
    def _(x):
        return x + 1

    @func.register(float)
    def _(x):
        return x + 100

    assert func(1) == 2
    assert func(1.0) == 101.0

    with pytest.raises(ValueError):
        func("s")

    @dispatch(argnum=1)
    def func(x, y):
        raise ValueError("unknown")

    @func.register(int)
    def _(x, y):
        return y + 1

    @func.register(float)
    def _(x, y):
        return y + 100

    assert func("a", 1) == 2
    assert func(None, 1.0) == 101.0

    with pytest.raises(ValueError):
        func(None, "s")

    # dispatch by keyword argument
    @dispatch(argnum="name")
    def func(x, y, *, name):
        raise ValueError("unknown")

    @func.register(int)
    def _(x, y, *, name):
        return "int"

    @func.register(str)
    def _(x, y, *, name):
        return "str"

    assert func(1, 2, name=1) == "int"
    assert func(1, 1, name="s") == "str"

    with pytest.raises(ValueError):
        func(1, 1, name=1.0)


def test_singledispatchmethod():
    class test:
        @dispatch(argnum=1)
        def plus(self, x):
            raise ValueError("unknown")

        @plus.register(int)
        def _(self, x):
            return x + 1

        @plus.register(complex)
        @plus.register(float)
        def _(self, x):
            return x + 100.0

    A = test()

    assert A.plus(1) == 2
    assert A.plus(1.0) == 101.0
    assert A.plus(complex(1.0)) == complex(101.0)

    with pytest.raises(ValueError):
        A.plus("s")

    with pytest.raises(ValueError):

        class test:
            @dispatch(argnum=1.0)
            def plus(self, x): ...

        t = test()
        t.plus("s")
        t.plus(1)

    @dispatch
    def fn(x):
        raise ValueError("unknown")

    @fn.register(int)
    def _(x):
        return 1

    assert fn(3) == 1

    with pytest.raises(ValueError):
        fn(3.0)



# adapted from https://github.com/deepmind/chex/blob/master/setup.py

import os

from setuptools import find_namespace_packages, setup

_CURRENT_DIR = os.path.dirname(os.path.abspath(__file__))


def _get_version():
    with open(os.path.join(_CURRENT_DIR, "pytreeclass", "__init__.py")) as fp:
        for line in fp:
            if line.startswith("__version__") and "=" in line:
                version = line[line.find("=") + 1 :].strip(" '\"\n")
                if version:
                    return version
        raise ValueError("`__version__` not defined in `pytreeclass/__init__.py`")


def _parse_requirements(path):

    with open(os.path.join(_CURRENT_DIR, path)) as f:
        return [
            line.rstrip() for line in f if not (line.isspace() or line.startswith("#"))
        ]


setup(
    name="pytreeclass",
    version=_get_version(),
    url="https://github.com/ASEM000/pytreeclass",
    license="MIT",
    author="Mahmoud Asem",
    description=("JAX compatible dataclass."),
    long_description=open(os.path.join(_CURRENT_DIR, "README.md")).read(),
    long_description_content_type="text/markdown",
    author_email="asem00@kaist.ac.kr",
    keywords="python machine-learning pytorch jax",
    packages=find_namespace_packages(exclude=['examples", "tests","experimental']),
    install_requires=_parse_requirements(
        os.path.join(_CURRENT_DIR, "requirements", "requirements.txt")
    ),
    # tests_require=_parse_requirements(
    #     os.path.join(_CURRENT_DIR, 'requirements', 'requirements-test.txt')),
    zip_safe=False,  # Required for full installation.
    python_requires=">=3.7",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
)



from __future__ import annotations

import inspect
import math
from types import FunctionType
from typing import Any, Sequence

import jax
import jax.numpy as jnp
import jax.tree_util as jtu
import jaxlib

from pytreeclass.src.decorator_util import dispatch

PyTree = Any


# Node formatting


def _func_repr(func):
    args, varargs, varkw, _, kwonlyargs, _, _ = inspect.getfullargspec(func)
    args = (",".join(args)) if len(args) > 0 else ""
    varargs = ("*" + varargs) if varargs is not None else ""
    kwonlyargs = (",".join(kwonlyargs)) if len(kwonlyargs) > 0 else ""
    varkw = ("**" + varkw) if varkw is not None else ""
    name = "Lambda" if (func.__name__ == "<lambda>") else func.__name__
    return (
        f"{name}("
        + ",".join(item for item in [args, varargs, kwonlyargs, varkw] if item != "")
        + ")"
    )


def jax_numpy_repr(node, *args, **kwargs):
    replace_tuple = (
        ("int", "i"),
        ("float", "f"),
        ("complex", "c"),
        (",)", ")"),
        ("(", "["),
        (")", "]"),
        (" ", ""),
    )

    formatted_string = f"{node.dtype}{jnp.shape(node)!r}"

    for lhs, rhs in replace_tuple:
        formatted_string = formatted_string.replace(lhs, rhs)
    return formatted_string


def _format_size(node_size, newline=False):
    """return formatted size from inexact(exact) complex number"""
    mark = "\n" if newline else ""
    order_kw = ["B", "KB", "MB", "GB"]

    # define order of magnitude
    real_size_order = int(math.log(node_size.real, 1024)) if node_size.real > 0 else 0
    imag_size_order = int(math.log(node_size.imag, 1024)) if node_size.imag > 0 else 0
    return (
        f"{(node_size.real)/(1024**real_size_order):.2f}{order_kw[real_size_order]}{mark}"
        f"({(node_size.imag)/(1024**imag_size_order):.2f}{order_kw[imag_size_order]})"
    )


def _format_count(node_count, newline=False):
    mark = "\n" if newline else ""
    return f"{int(node_count.real):,}{mark}({int(node_count.imag):,})"


def _format_node_repr(node, depth):
    @dispatch(argnum=0)
    def __format_node_repr(node, *args, **kwargs):
        multiline = "\n" in f"{node!s}"
        string = ("\n" + "\t" * (depth + 1)) if multiline else ""
        string += ("\n" + "\t" * (depth + 1)).join(f"{node!r}".split("\n"))
        return string

    @__format_node_repr.register(jnp.ndarray)
    @__format_node_repr.register(jax.ShapeDtypeStruct)
    def _(node, *args, **kwargs):
        return jax_numpy_repr(node)

    @__format_node_repr.register(jaxlib.xla_extension.CompiledFunction)
    @__format_node_repr.register(jax._src.custom_derivatives.custom_jvp)
    @__format_node_repr.register(FunctionType)
    def _(node, *args, **kwargs):
        return _func_repr(node)

    @__format_node_repr.register(list)
    def _(node, depth):
        string = ",".join(_format_node_repr(layer, depth=depth) for layer in node)
        shifted = "\t" * (depth) + string
        return "[" + shifted + "]"

    @__format_node_repr.register(tuple)
    def _(node, depth):
        string = ",".join(_format_node_repr(layer, depth=depth) for layer in node)
        shifted = "\t" * (depth) + string
        return "(" + shifted + ")"

    @__format_node_repr.register(dict)
    def _(node, depth):
        string = ",".join(
            f"{k}:{_format_node_repr(v,depth=depth)}" for k, v in node.items()
        )
        shifted = "\t" * (depth) + string
        return "{" + shifted + "}"

    return __format_node_repr(node, depth)


def _format_node_str(node, depth):
    @dispatch(argnum=0)
    def __format_node_str(node, depth):
        multiline = "\n" in f"{node!s}"
        string = ("\n" + "\t" * (depth + 1)) if multiline else ""
        string += ("\n" + "\t" * (depth + 1)).join(f"{node!s}".split("\n"))
        return string

    @__format_node_str.register(jaxlib.xla_extension.CompiledFunction)
    @__format_node_str.register(jax._src.custom_derivatives.custom_jvp)
    @__format_node_str.register(FunctionType)
    def _(node, *args, **kwargs):
        return _func_repr(node)

    @__format_node_str.register(list)
    def _(node, depth):
        string = ",".join(_format_node_str(layer, depth=depth) for layer in node)
        shifted = "\t" * (depth) + string
        return "[" + shifted + "]"

    @__format_node_str.register(tuple)
    def _(node, depth):
        string = ",".join(_format_node_str(layer, depth=depth) for layer in node)
        shifted = "\t" * (depth) + string
        return "(" + shifted + ")"

    @__format_node_str.register(dict)
    def _(node, depth):
        string = ",".join(
            f"{k}:{_format_node_str(v,depth=depth)}" for k, v in node.items()
        )
        shifted = "\t" * (depth) + string
        return "{" + shifted + "}"

    return __format_node_str(node, depth)


def _format_node_diagram(node, *args, **kwargs):
    @dispatch(argnum=0)
    def __format_node_diagram(node, *args, **kwargs):
        return f"{node!r}"

    @__format_node_diagram.register(jaxlib.xla_extension.CompiledFunction)
    @__format_node_diagram.register(jax._src.custom_derivatives.custom_jvp)
    @__format_node_diagram.register(FunctionType)
    def _(node, *args, **kwargs):
        return _func_repr(node)

    @__format_node_diagram.register(jnp.ndarray)
    @__format_node_diagram.register(jax.ShapeDtypeStruct)
    def _(node, *args, **kwargs):
        return jax_numpy_repr(node)

    return __format_node_diagram(node, *args, **kwargs)


# Box drawing


def _hbox(*text):

    boxes = list(map(_vbox, text))
    boxes = [(box).split("\n") for box in boxes]
    max_col_height = max([len(b) for b in boxes])
    boxes = [b + [" " * len(b[0])] * (max_col_height - len(b)) for b in boxes]
    FMT = ""

    for _, line in enumerate(zip(*boxes)):
        FMT += _resolve_line(line) + "\n"
    return FMT


def _hstack(boxes):

    boxes = [(box).split("\n") for box in boxes]
    max_col_height = max([len(b) for b in boxes])

    # expand height of each col before merging
    boxes = [b + [" " * len(b[0])] * (max_col_height - len(b)) for b in boxes]

    FMT = ""

    _cells = tuple(zip(*boxes))

    for i, line in enumerate(_cells):
        FMT += _resolve_line(line) + ("\n" if i != (len(_cells) - 1) else "")

    return FMT


def _vbox(*text: tuple[str, ...]) -> str:
    """Create vertically stacked text boxes

    Returns:
        str: stacked boxes string

    Examples:
    >>> _vbox("a","b")
        ┌───┐
        │a  │
        ├───┤
        │b  │
        └───┘

    >>> _vbox("a","","a")
        ┌───┐
        │a  │
        ├───┤
        │   │
        ├───┤
        │a  │
        └───┘
    """

    max_width = (
        max(jtu.tree_flatten([[len(t) for t in item.split("\n")] for item in text])[0])
        + 0
    )

    top = f"┌{'─'*max_width}┐"
    line = f"├{'─'*max_width}┤"
    side = [
        "\n".join([f"│{t}{' '*(max_width-len(t))}│" for t in item.split("\n")])
        for item in text
    ]
    btm = f"└{'─'*max_width}┘"

    formatted = ""

    for i, s in enumerate(side):

        if i == 0:
            formatted += f"{top}\n{s}\n{line if len(side)>1 else btm}"

        elif i == len(side) - 1:
            formatted += f"\n{s}\n{btm}"

        else:
            formatted += f"\n{s}\n{line}"

    return formatted


def _resolve_line(cols: Sequence[str, ...]) -> str:
    """combine columns of single line by merging their borders

    Args:
        cols (Sequence[str,...]): Sequence of single line column string

    Returns:
        str: resolved column string

    Example:
        >>> _resolve_line(['ab','b│','│c'])
        'abb│c'

        >>> _resolve_line(['ab','b┐','┌c'])
        'abb┬c'

    """

    cols = list(map(list, cols))  # convert each col to col of chars
    alpha = ["│", "┌", "┐", "└", "┘", "┤", "├"]

    for index in range(len(cols) - 1):

        if cols[index][-1] == "┐" and cols[index + 1][0] in ["┌", "─"]:
            cols[index][-1] = "┬"
            cols[index + 1].pop(0)

        elif cols[index][-1] == "┘" and cols[index + 1][0] in ["└", "─"]:
            cols[index][-1] = "┴"
            cols[index + 1].pop(0)

        elif cols[index][-1] == "┤" and cols[index + 1][0] in ["├", "─", "└"]:  #
            cols[index][-1] = "┼"
            cols[index + 1].pop(0)

        elif cols[index][-1] in ["┘", "┐", "─"] and cols[index + 1][0] in ["├"]:
            cols[index][-1] = "┼"
            cols[index + 1].pop(0)

        elif cols[index][-1] == "─" and cols[index + 1][0] == "└":
            cols[index][-1] = "┴"
            cols[index + 1].pop(0)

        elif cols[index][-1] == "─" and cols[index + 1][0] == "┌":
            cols[index][-1] = "┬"
            cols[index + 1].pop(0)

        elif cols[index][-1] == "│" and cols[index + 1][0] == "─":
            cols[index][-1] = "├"
            cols[index + 1].pop(0)

        elif cols[index][-1] == " ":
            cols[index].pop()

        elif cols[index][-1] in alpha and cols[index + 1][0] in [*alpha, " "]:
            cols[index + 1].pop(0)

    return "".join(map(lambda x: "".join(x), cols))


def _table(lines):
    """

    === Explanation
        create a _table with self aligning rows and cols

    === Args
        lines : list of lists of cols values

    === Examples
        >>> print(_table([['1\n','2'],['3','4000']]))
            ┌─┬────────┐
            │1│3       │
            │ │        │
            ├─┼────────┤
            │2│40000000│
            └─┴────────┘


    """
    # align _cells vertically
    for i, _cells in enumerate(zip(*lines)):
        max__cell_height = max(map(lambda x: x.count("\n"), _cells))
        for j in range(len(_cells)):
            lines[j][i] += "\n" * (max__cell_height - lines[j][i].count("\n"))
    cols = [_vbox(*col) for col in lines]

    return _hstack(cols)


def _layer_box(name, indim=None, outdim=None):
    """
    === Explanation
        create a keras-like layer diagram

    ==== Examples
        >>> print(_layer_box("Test",(1,1,1),(1,1,1)))
        ┌──────┬────────┬───────────┐
        │      │ Input  │ (1, 1, 1) │
        │ Test │────────┼───────────┤
        │      │ Output │ (1, 1, 1) │
        └──────┴────────┴───────────┘

    """

    return _hstack(
        [
            _vbox(f"\n {name} \n"),
            _table([[" Input ", " Output "], [f" {indim} ", f" {outdim} "]]),
        ]
    )



import jax.numpy as jnp
import pytest

from pytreeclass import treeclass
from pytreeclass.src.tree_util import (
    _node_count_and_size,
    is_treeclass,
    is_treeclass_leaf,
    _dispatched_tree_map
)


@treeclass
class Test:
    a: int = 10


a = Test()
b = (1, "s", 1.0, [2, 3])


@treeclass
class Test2:
    a: int = 1
    b: Test = Test()


def test_is_treeclass():
    assert is_treeclass(a) is True
    assert all(is_treeclass(bi) for bi in b) is False


def test_is_treeclass_leaf():
    assert is_treeclass_leaf(a) is True
    assert all(is_treeclass_leaf(bi) for bi in b) is False
    assert is_treeclass_leaf(Test2()) is False
    assert is_treeclass_leaf(Test2().b) is True


def test__node_count_and_size():
    @treeclass
    class Test:
        a: jnp.ndarray = jnp.array([1.0, 2.0, 3.0])
        b: int = 1

    t = Test()
    assert _node_count_and_size(t.b) == (complex(0, 1), complex(0, 28))
    assert _node_count_and_size(t.a) == (complex(3, 0), complex(12, 0))

    assert _node_count_and_size(jnp.array([1, 2, 3, 4, 5])) == (
        complex(0, 5),
        complex(0, 20),
    )
    assert _node_count_and_size(3.0) == (complex(1), complex(24))

    @treeclass
    class x:
        a: int = 1
        b: int = 2

    assert x().treeclass_leaves == [x()]

    assert hash(x())

    xx = x()
    xx.cc = 1
    assert xx.cc == 1
    xx = xx.freeze()

    with pytest.raises(ValueError):
        xx.test = 1

    
def test_dispatched_tree_map():

    with pytest.raises(NotImplementedError):
        class A : ... 
        _dispatched_tree_map(lambda x,y: x, 1,A())




from dataclasses import field

import jax
import jax.tree_util as jtu
import numpy.testing as npt
import pytest
from jax import numpy as jnp

import pytreeclass as pytc
from pytreeclass.src.tree_util import static_value


def test_jit_freeze():
    @pytc.treeclass
    class Linear:
        weight: jnp.ndarray
        bias: jnp.ndarray
        name: str = field(default=static_value("a"))

        def __init__(self, key, in_dim, out_dim):
            self.weight = jax.random.normal(key, shape=(in_dim, out_dim)) * jnp.sqrt(
                2 / in_dim
            )
            self.bias = jnp.ones((1, out_dim))

        def __call__(self, x):
            return x @ self.weight + self.bias

    @pytc.treeclass
    class StackedLinear:
        l1: Linear
        l2: Linear
        l3: Linear

        def __init__(self, key, in_dim, out_dim, hidden_dim):
            keys = jax.random.split(key, 3)
            self.l1 = Linear(key=keys[0], in_dim=in_dim, out_dim=hidden_dim)
            self.l2 = Linear(key=keys[1], in_dim=hidden_dim, out_dim=hidden_dim)
            self.l3 = Linear(key=keys[2], in_dim=hidden_dim, out_dim=out_dim)

        def __call__(self, x):
            x = self.l1(x)
            x = jax.nn.tanh(x)
            x = self.l2(x)
            x = jax.nn.tanh(x)
            x = self.l3(x)
            return x

    model = StackedLinear(in_dim=1, out_dim=1, hidden_dim=10, key=jax.random.PRNGKey(0))
    x = jnp.linspace(0, 1, 100)[:, None]
    y = x**3 + jax.random.uniform(jax.random.PRNGKey(0), (100, 1)) * 0.01

    @jax.value_and_grad
    def loss_func(model, x, y):
        return jnp.mean((model(x) - y) ** 2)

    @jax.jit
    def update(model, x, y):
        value, grads = loss_func(model, x, y)
        return value, jtu.tree_map(lambda x, y: x - 1e-3 * y, model, grads)

    # freeze l1
    def train_step(x, y, epochs=20_000):
        model = StackedLinear(
            in_dim=1, out_dim=1, hidden_dim=10, key=jax.random.PRNGKey(0)
        )
        model.l1 = model.l1.freeze()
        for i in range(1, epochs + 1):
            value, model = update(model, x, y)

        npt.assert_allclose(value, jnp.array(0.0012702086), atol=1e-4)
        return value, model

    for _ in range(2):
        value, model = train_step(x, y, epochs=20_000)

    # freeze l2
    def train_step(x, y, epochs=20_000):
        model = StackedLinear(
            in_dim=1, out_dim=1, hidden_dim=10, key=jax.random.PRNGKey(0)
        )
        model.l2 = model.l2.freeze()
        for i in range(1, epochs + 1):
            value, model = update(model, x, y)

        npt.assert_allclose(value, jnp.array(0.00619382), atol=1e-4)
        return value, model

    for _ in range(2):
        value, model = train_step(x, y, epochs=20_000)

    # freeze all
    def train_step(x, y, epochs=20_000):
        model = StackedLinear(
            in_dim=1, out_dim=1, hidden_dim=10, key=jax.random.PRNGKey(0)
        )
        model = model.freeze()
        for i in range(1, epochs + 1):
            value, model = update(model, x, y)

        npt.assert_allclose(value, jnp.array(3.9368904), atol=1e-4)
        return value, model

    for _ in range(2):
        value, model = train_step(x, y, epochs=20_000)


def test_ops_with_jit():
    @pytc.treeclass
    class T0:
        a: int = 1
        b: int = 2
        c: int = 3

    @pytc.treeclass
    class T1:
        a: int = 1
        b: int = 2
        c: int = 3
        d: jnp.ndarray = jnp.array([1, 2, 3])

    @jax.jit
    def getter(tree):
        return tree.at[...].get()

    @jax.jit
    def setter(tree):
        return tree.at[...].set(0)

    @jax.jit
    def applier(tree):
        return tree.at[...].apply(lambda _: 0)

    with pytest.raises(jax.errors.ConcretizationTypeError):
        getter(T0())

    assert pytc.src.tree_util.is_treeclass_equal(T0(0, 0, 0), setter(T0()))

    assert pytc.src.tree_util.is_treeclass_equal(T0(0, 0, 0), applier(T0()))

    with pytest.raises(jax.errors.ConcretizationTypeError):
        getter(T1())

    assert pytc.src.tree_util.is_treeclass_equal(
        T1(0, 0, 0, jnp.array([0, 0, 0])), setter(T1())
    )

    assert pytc.src.tree_util.is_treeclass_equal(
        T1(0, 0, 0, jnp.array([0, 0, 0])), applier(T1())
    )
